package typesampling;

import java.util.*;
import fig.basic.*;
import fig.prob.*;
import fig.exec.*;
import static fig.basic.LogInfo.*;

/**
 * Concentration parameter for the Dirichlet process ($\alpha_0$).
 * Use the Mike West trick to sample it using data augmentation.
 * Details are on page 4 of:
 *   http://www.cs.berkeley.edu/~jordan/sail/readings/west.ps
 * $p(\alpha, x, K) = p(\alpha) p(K | \alpha)$
 * Algorithm: alternate between
 *  - sample $p(x | \alpha, K)$
 *  - sample $p(\alpha | x, K)$
 */
public class Concentration implements DeepCloneable<Concentration> {
  private Gamma prior; // Specifies $p(\alpha)$
  private double alpha; // Actual variable we want to sample
  private double x; // Auxiliary variable to faciliate sampling $\alpha$
  private int n; // Number of data points
  private int K; // Number of clusters
  private double sumAlpha = 0;

  // Number of mini-steps to do per sample of $\alpha$:
  // only one is necessary for ergodicity
  private int numMiniSteps = 5;

  private Concentration() { }

  public Concentration(double concentrationPriorRate, int n) {
    this.prior = new Gamma(1, concentrationPriorRate);
    this.alpha = prior.getMean(); // Initial value
    this.n = n; // Fixed
    this.x = 0.5;
    this.K = -1;
  }

  public double getAlpha() { return alpha; }
  public double logProb() { return prior.logProb(alpha); }

  public Concentration deepClone() {
    Concentration concentration = new Concentration();
    concentration.prior = this.prior;
    concentration.alpha = this.alpha;
    concentration.x = this.x;
    concentration.n = this.n;
    concentration.K = this.K;
    concentration.sumAlpha = this.sumAlpha;
    return concentration;
  }

  public void doSampleStep(Random random, int K) {
    if (n == 0) {
      assert K == 0;
      this.alpha = prior.sample(random);
    }
    else {
      this.K = K;
      // Sample from $p(\alpha, x | K)$
      for(int i = 0; i < numMiniSteps; i++) {
        doSampleX(random);
        doSampleAlpha(random);
        //logs("alpha=%f (avg %f), x=%f, K=%d", alpha, sumAlpha/(i+1), x, K);
      }
    }
  }

  private void doSampleX(Random random) {
    this.x = Beta.sample(random, alpha+1, n);
    //logs("Beta(%f, %d) = %f", alpha+1, n, x);
  }

  private void doSampleAlpha(Random random) {
    // Sample from a mixture of two gammas
    double A = prior.getShape() + K - 1;
    double B = prior.getRate() - Math.log(x);

    double y = A / (n * B);
    double pi = y / (1 + y);
    if(Binomial.sample(random, pi)) A += 1;

    //logs("A = %f, B = %f", A, B);

    this.alpha = Gamma.sample(random, A, B);
    sumAlpha += alpha;
  }

  public String toString() { return Fmt.D(alpha); }
}
