package typesampling

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid}
import java.util.{Set,ArrayList,Random}

/**
A counter is a map from keys to counts.
Allows fast access to sum of counts.
These counts form a distribution from which it's easy to sample.
*/

trait Counter[A] {
  // Add/remove
  def incr(a:A, dv:Double) : Unit
  def incr(scale:Double, that:Counter[A]) : Unit = {
    that.keys.foreach { a:A =>
      incr(a, scale*that.getCount(a))
    }
  }
  def remove(a:A) : Unit
  def clear : Unit

  // Fetch information
  def getCount(a:A) : Double
  def getProb(a:A) : Double = getCount(a) / totalCount
  def totalCount : Double
  def keys : Iterator[A]
  def numKeys : Int
  def numNonZero : Int

  def argmaxAndMax : (Option[A],Double) = {
    var bestKey : Option[A] = None
    var bestValue = 0.0
    keys.foreach { key =>
      val value = getCount(key)
      if (bestKey == None || value > bestValue) {
        bestKey = Some(key)
        bestValue = value
      }
    }
    (bestKey, bestValue)
  }

  def sample(random:Random) : A = {
    val target = random.nextDouble * totalCount
    val iterator = keys
    def recurse(accum:Double) : A = {
      val x = iterator.next
      val newAccum = accum + getCount(x)
      if (newAccum > target) x // Choose
      else recurse(newAccum)
    }
    recurse(0)
  }

  def foreachSorted(numTop:Int, f:(A,Double) => Any) = {
    val keys = new Array[Any](numKeys)
    this.keys.copyToArray(keys, 0)
    Utils.foreachSorted(keys.map{k => getCount(k.asInstanceOf[A])}, numTop, true, { (i:Int,v:Double) =>
      f(keys(i).asInstanceOf[A], v)
    })
  }

  def write(numTop:Int, alsoPrintNormalized:Boolean, prefix:String, puts:String=>Any) = {
    foreachSorted(numTop, { (a:A,v:Double) =>
      puts(prefix + a + "\t" + fmt(v) + (if (alsoPrintNormalized) " "+fmt(v/totalCount) else ""))
    })
  }

  def removeEntriesWithSmallCounts(threshold:Double) = {
    keys.foreach { a:A =>
      if (getCount(a) < threshold) remove(a)
    }
  }
}

////////////////////////////////////////////////////////////

class CounterMap[A] extends Counter[A] {
  // Note: don't store 0 counts
  private val values = new HashMap[A,Double]
  private var sumValue = 0.0

  def incr(a:A, dv:Double) : Unit = {
    val v = values.getOrElse(a, 0.0) + dv
    sumValue += dv
    if (v == 0) values -= a
    else values(a) = v
    //require (v >= 0)
  }

  def remove(a:A) = { values -= a }
  def clear = {
    values.clear
    sumValue = 0.0
  }

  def getCount(a:A) = values.getOrElse(a, 0.0)
  def totalCount = sumValue
  def keys = values.keys.iterator
  def numKeys = values.size
  def numNonZero = values.size
}

////////////////////////////////////////////////////////////

class CounterVec(N:Int) extends Counter[Int] {
  // Keys are 0..N-1
  private val values = new Array[Double](N)
  private var sumValue = 0.0

  def incr(a:Int, dv:Double) : Unit = {
    values(a) += dv
    sumValue += dv
  }

  def remove(a:Int) = { }
  def clear = {
    foreach(N, { a:Int => values(a) = 0 })
    sumValue = 0.0
  }

  def getCount(a:Int) = values(a)
  def totalCount = sumValue
  def keys = (0 to N-1).iterator
  def numKeys = N
  def numNonZero = Utils.count(N, { i:Int => values(i) != 0 })
}

////////////////////////////////////////////////////////////

// A map from some key to a counter (for conditional probabilities).
abstract class CounterCollection[A,B] {
  private val maps = new HashMap[A, Counter[B]]

  def newCounter(a:A) : Counter[B]

  def incr(a:A, b:B, dv:Double) : Unit = {
    val map = maps.getOrElseUpdate(a, newCounter(a))
    map.incr(b, dv)
    if (map.numNonZero == 0) maps -= a
  }
  def incr(a:A, scale:Double, that:Counter[B]) : Unit = {
    val map = maps.getOrElseUpdate(a, newCounter(a))
    map.incr(scale, that)
  }
    
  def incr(that:CounterCollection[A,B]) = {
    that.keys.foreach { a:A =>
      val map = maps.getOrElseUpdate(a, newCounter(a))
      map.incr(1, that(a))
    }
  }

  def foreachSorted(numTop:Int, f:(A,Counter[B]) => Any) = {
    val keys = new Array[Any](maps.size)
    maps.keys.copyToArray(keys, 0)
    Utils.foreachSorted(keys.map{k => maps(k.asInstanceOf[A]).totalCount}, numTop, true, { (i:Int,v:Double) =>
      f(keys(i).asInstanceOf[A], maps(keys(i).asInstanceOf[A]))
    })
  }
  def write(numTopA:Int, numTopB:Int, alsoPrintNormalized:Boolean, prefix:String, puts:String=>Any) = {
    foreachSorted(numTopA, { (a:A,map:Counter[B]) =>
      map.write(numTopB, alsoPrintNormalized, prefix+a+"\t", puts)
    })
  }

  def sample(random:Random, a:A) : B = {
    maps.get(a) match {
      case None => throw new RuntimeException("No distribution for " + a)
      case Some(map) => map.sample(random)
    }
  }

  def getProb(a:A, b:B) = {
    maps.get(a) match {
      case None => 0.0
      case Some(map) => map.getProb(b)
    }
  }
  def getCount(a:A, b:B) = {
    maps.get(a) match {
      case None => 0.0
      case Some(map) => map.getCount(b)
    }
  }
  def totalCount(a:A) = {
    maps.get(a) match {
      case None => 0.0
      case Some(map) => map.totalCount
    }
  }

  def numKeys = maps.size
  def keys = maps.keys
  def values = maps.values
  def apply(a:A) = maps(a)
  def get(a:A) = maps.get(a)

  def numNonZero = {
    var n = 0
    maps.values.foreach(n += _.numNonZero)
    n
  }
}

class CounterMapCollection[A,B] extends CounterCollection[A,B] {
  def newCounter(a:A) = new CounterMap[B]
}

class CounterVecCollection[A](K:Int) extends CounterCollection[A,Int] {
  def newCounter(a:A) = new CounterVec(K)
}
