package typesampling.flatseg2

import typesampling._
import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import fig.exec.Execution.putLogRec
import fig.basic.NumUtils.logGamma
import Misc.logGenFact

////////////////////////////////////////////////////////////

// Given a sequence of atoms (strings), group them into segments
case class FlatSegLearner2(opts:LearnOptions) extends typesampling.TypeBasedLearner with typesampling.flatseg.FlatSegBase {
  trait Pattern
  case class Pattern1(s:Seg) extends Pattern
  case class Pattern2(s1:Seg, s2:Seg) extends Pattern

  // A site looks like this:
  // ex:              i
  //             ... a?b ...
  // where ? could either but cut or not.
  // prev and next could be computed, but we pass it in 
  case class Site(ex:Example, i:Int, prev:Int, next:Int) extends AbstractSite {
    require (i > 0 && i < ex.N)

    def sameAs_?(that:Site) = (this.ex eq that.ex) && this.i == that.i

    def removeNeighboringSpotsPermanently = {
                       ex.removeSpotPermanently(i)
                       ex.removeSpotPermanently(prev)
      if (next < ex.N) ex.removeSpotPermanently(next)
    }
    def removeNeighboringSpotsTemporarily = {
      if (ex.cut(i))   ex.removeSpotTemporarily(i, true)
                       ex.removeSpotTemporarily(prev, false)
      if (next < ex.N) ex.removeSpotTemporarily(next, false)
    }
    def addNeighboringSpots = {
      if (ex.cut(i))   ex.addSpot(i, true)
                       ex.addSpot(prev, false)
      if (next < ex.N) ex.addSpot(next, false)
    }

    override def toString = "("+i+")"
  }

  // A spot occurs at and only at cut points (cut(i) = true).
  case class Spot(ex:Example, i:Int) extends AbstractSpot {
    require (ex.cut(i), i)

    val occurrences = {
      if (i == 0)
        PatternOccurrence(Pattern1(ex.rightSeg(i)), this) :: Nil
      else
        PatternOccurrence(Pattern1(ex.rightSeg(i)), this) ::
        PatternOccurrence(Pattern2(ex.leftSeg(i), ex.rightSeg(i)), this) :: Nil
    }

    override def toString = "["+i+"]"
  }

  // Contains elements, predicted segmentation cut, and true segmentation trueCut
  // cut(i) = whether atom i start the beginning of a new segment
  class Example(atoms:Array[String], cut:Array[Boolean], trueCut:Array[Boolean]) extends ExampleBase(atoms, cut, trueCut) with SpotSeq {
    val spots : Array[Option[Spot]] = map(N, { i:Int => None })
    def newSpot(i:Int) = Spot(this, i)
    def initializeSpots = foreach(N, { i:Int => if (cut(i)) addSpot(i, true) })
    def foreachSite(f:Site=>Any) = foreach(1, N, { i:Int =>
      if (!opts.sampleSiteOncePerIter || iter != lastIterSampled(i))
        f(Site(this, i, compute_prev(i), compute_next(i)))
    })
  }
  def newExample(atoms:Array[String], cut:Array[Boolean], trueCut:Array[Boolean]) = new Example(atoms, cut, trueCut)

  override def learnInit = {
    super.learnInit
    Utils.track_foreach("Example", trainExamples, { ex:Example =>
      foreach(ex.N, { i:Int => params.atomCounts.incr(ex.atoms(i), 1) })
      true
    })
  }

  ////////////////////////////////////////////////////////////

  class IterState extends super.IterState {
    override def log = {
      super.log
      val numSegs = Utils.sum(trainExamples.size, { i:Int => trainExamples(i).numSegs })
      putLogRec("totalSegs", numSegs)
      putLogRec("avgSegs", 1.0*numSegs/trainExamples.size)
    }

    override def sampleHyperparameters = {
      if (opts.sampleAlpha)
        params.alpha = sampleAlpha(params.segCounts)
    }

    def setCut(ex:Example, i:Int, newCut:Boolean) : Boolean = {
      numSampledVars += 1
      ex.lastIterSampled(i) = iter
      if (ex.cut(i) != newCut) {
        ex.setCut(i, newCut)
        numChangedVars += 1
        true
      }
      else
        false
    }

    def newMoveState(site:Site) = Some(MoveState(site))
    case class MoveState(origSite:Site) extends super.MoveState {
      val s1 = new Seg(origSite.ex.atoms, origSite.prev, origSite.i)
      val s2 = new Seg(origSite.ex.atoms, origSite.i, origSite.next)
      val s3 = new Seg(origSite.ex.atoms, origSite.prev, origSite.next)
      val dup = s1 == s2
      val patterns = Pattern1(s3) :: Pattern2(s1, s2) :: Nil

      def incrCounts(k:Int, d:Double) = {
        if (k == 0)
          params.segCounts.incr(s3, d)
        else if (k == 1) {
          params.segCounts.incr(s1, d)
          params.segCounts.incr(s2, d)
        }
        else
          throw impossible
      }

      def getSite(o:PatternOccurrence) = {
        val ex = o.spot.ex
        val i = o.spot.i
        o.pattern match {
          case pattern:Pattern1 =>
            Some(Site(ex, i + s1.size, i, i + s3.size))
          case pattern:Pattern2 =>
            ex.returnIfSpotExists(i, Site(ex, i, i-s1.size, i+s2.size))
        }
      }

      def numConfigurations = 2
      def getConfiguration(site:Site) = if (site.ex.cut(site.i)) 1 else 0
      def setConfiguration(site:Site, c:Int) = setCut(site.ex, site.i, c == 1)

      def newScorer = new Scorer
      class Scorer extends AbstractScorer {
        val b1 = params.alpha*baseProb(s1) + params.segCounts.getCount(s1)
        val b2 = params.alpha*baseProb(s2) + params.segCounts.getCount(s2)
        val b3 = params.alpha*baseProb(s3) + params.segCounts.getCount(s3)
        val B  = params.alpha              + params.segCounts.totalCount
        if (opts.verbose >= 5) dbgs("Scorer: b1=%s, b2=%s, b3=%s, B=%s", b1, b2, b3, B)

        val logContProb = math.log(params.segContProb)

        def indepScore(c:Int) = {
          if (c == 0) { // Not cut
            logGenFact(b3, N) - logGenFact(B, N) // ab or aa
          }
          else if (c == 1) { // Cut
            if (dup) // aa
              logGenFact(b1, 2*N) - logGenFact(B, 2*N) + N*logContProb
            else // ab
              logGenFact(b1, N) + logGenFact(b2, N) - logGenFact(B, 2*N) + N*logContProb
          }
          else
            throw impossible
        }

        override def localFeatures(k:Int, n:Int) = {
          require (k == 2)
          //Array(k, n, b1, b2, b3)
          Array(1.0, math.log(b1), math.log(b2), math.log(b3))
        }

        def localScore(k:Int, n:Int, m:Int) = {
          val c = k-1
          if (c == 0) { // Not cut
            logGenFact(b3, m) // ab or aa
          }
          else if (c == 1) { // Cut
            if (dup) // aa
              logGenFact(b1, 2*m) - logGenFact(B, N+m) + m*logContProb
            else // ab
              logGenFact(b1, m) + logGenFact(b2, m) - logGenFact(B, N+m) + m*logContProb
          }
          else
            throw impossible
        }

        override def computeScores(k:Int, n:Int) = {
          require (k == 2) 
          val scores = new Array[Double](n+1)
          var sum = 0.0
          foreach(1, n+1, { m:Int => // m is the number of cuts
            // Order of operations is important (m-1+b1 != b1+m-1 when m = 1 and b1 = 1e-30)
            sum += math.log(((b1+(m-1)) * (b2+(m-1)) / (B+(n+m-1)) / (b3+(n-m))) * (n-m+1) / m) + logContProb
            scores(m) = sum
          })
          scores
        }
      }
    }
  }
  def newIterState = new IterState

  override def repCheck = {
    super.repCheck
    trainExamples.foreach { ex:Example =>
      /*dbgs("prev: %s", ex.prev.mkString(" "))
      dbgs("next: %s", ex.next.mkString(" "))
      dbgs("spots: %s", ex.spots.mkString(" "))
      dbgs("cut: %s", ex.cut.mkString(" "))*/
      foreach(ex.N+1, { i:Int =>
        if (ex.cut(i)) {
          require (ex.prev(i) == (if (i == 0) -1 else ex.compute_prev(i)))
          require (ex.next(i) == (if (i == ex.N) -1 else ex.compute_next(i)))
        }
        else {
          require (ex.prev(i) == -1)
          require (ex.next(i) == -1)
        }
      })
    }
  }
}
