package typesampling

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,fail,impossible}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import fig.exec.Execution.putLogRec
import fig.basic.NumUtils.logGamma
import fig.record.Record

class LearnOptions {
  import tea.OptionTypes._
  @Option var verbose = 0
  @Option var repCheck = 0
  @Option var paramsMaxPrint = Integer.MAX_VALUE

  @Option var maxIters = 5
  @Option var burnInFrac = 0.0
  @Option var learnerType = Enums.LearnerType.flatseg
  @Option var initType = Enums.InitType.random

  @Option var typeLevelSampling = false
  @Option var multiType = false

  @Option(gloss="For annealing") var initTemperature = 1.0
  @Option(gloss="For annealing") var finalTemperature = 1.0

  // For flat segmentation
  @Option(gloss="p(x) \\propto e^{-|x|^\\beta") var useDoubleExpPrior = false
  @Option(gloss="\\beta") var doubleExpPriorRate = 1.5
  @Option(gloss="How often to use dynamic program") var dynamicProgramFrac = 0.0
  @Option(gloss="When using a dynamic program") var maxSegmentLength = 5

  // For HMMs
  @Option var numStates = 4

  @Option var sampleAlpha = false
  @Option var sampleBaseContProb = false
  @Option(gloss="alpha ~ Gamma(1, alphaPriorRate)") var alphaPriorRate = 1.0
  @Option(gloss="Probability of joining a new table") var alpha = 1.0
  @Option(gloss="Probability of continuing the tree fragment/segment (also used as base measure)") var initBaseContProb = 0.5
  @Option var initRandom = new Random(1)
  @Option var sampleRandom = new Random(1)

  @Option(gloss="When initialize randomly (1 = small trees, 0 = one large tree)") var initCutProb = 0.0

  // Debugging output
  @Option var writeInferenceStatsInterval = 1
  @Option var writeInferenceStats = false
  @Option var writePredictions = true
  @Option var accumulateFullStats = false
  @Option var printContext = true

  // For the PTSG model
  @Option var numRefactorCandidatesToApply = 1
  @Option var refactorMinArgLists = 2
  @Option(gloss="Maximum size of contexts (order of functions) that we can allow as arguments") var maxArgContextSize = 0

  // Parsing
  @Option(gloss="Whether to evaluate learned model on WSJ parsing") var evalParsing = false
  @Option(gloss="Prefix of files for evaluating WSJ parsing") var evalParsingDatasetPrefix = "eval-parsing/datasets/dev"

  @Option var posteriorThreshold = 0.5
  @Option var usePosteriorDecoding = false

  // For approximate type-sampling
  @Option var trainProposal = false
  @Option var proposalInitStepSize = 1.0
  @Option var proposalStepSizeReductionPower = 0.5
  @Option var useProposal = false
  @Option var hardMoves = false
  @Option var considerOnlyVertices = false
  @Option var alwaysAccept = false
  @Option var polyaConcentration = 1.0
  @Option var polyaSmoothing = 1e-4
  @Option var hmmConsiderStatePairs = false
  @Option var sampleSiteOncePerIter = false

  @Option var proposalHeldoutFrac = 1e-5
  @Option var proposalPrecision = 1e-4
  @Option var gibbsOnSimplex = false
  @Option var investigateScoresShape = false

  // For EM
  @Option var em = false
  @Option var uniformInitParams = false

  @Option var hmmUseTaggingDictionary = false

  // Lexical model
  //@Option var lexicalModelType = Enums.LexicalModelType.uniform
  //@Option var newLexProb = 1e-5
}

trait Learner {
  type Example <: {
    def log : Unit
    def write(puts:String=>Any) : Unit
  }
  type Params <: {
    def write(maxPrint:Int, puts:String=>Any) : Unit
    def writeGrammar(puts:String=>Any) : Unit
    def size : Int // Number of parameters
  }

  trait AbstractParams {
    def writeGrammar(puts:String=>Any) : Unit = throw new RuntimeException("Not implemented")
  }

  def convertNode(node:Node) : Example

  var trainExamples : Seq[Example] = null
  var testExamples : Seq[Example] = null
  val lexicon = new Lexicon // For parsing: abstract away rare words

  def setTrainNodes(trainNodes:Array[Node]) = {
    if (opts.evalParsing)
      lexicon.buildLexicon(trainNodes)
    this.trainExamples = trainNodes.map { _node:Node =>
      var node = _node
      if (opts.evalParsing) node = lexicon.abstractRareWords(node)
      convertNode(node)
    }
  }

  def setTestNodes(testNodes:Array[Node]) = { this.testExamples = testNodes.map(convertNode(_)) }

  var isIntermediateLearner = false // If intermediate, don't print out stuff
  val opts : LearnOptions

  def stateLogProb : Double
  def stateAccuracy = Double.NaN
  def logProb(node:Node) = Double.NaN

  def stateStr(ex:Example) : String = throw fail("not implemented")
  def getInferenceStats(ex:Example) : Array[Double]
  def writeInferenceStats(path:String) = {
    Utils.writeLines(path, { puts:(String=>Any) =>
      trainExamples.foreach { ex:Example =>
        puts(fmt1(getInferenceStats(ex)))
      }
    })
  }

  def repCheck : Unit
  private def track_repCheck = track("repCheck") { repCheck }

  val params : Params
  val avgParams : Params

  def log = {
    putLogRec("numParams", params.size)
    putLogRec("logZ", stateLogProb)
    putLogRec("accuracy", stateAccuracy)
  }

  // Ways of measuring time
  var iter = 0
  val timer = new fig.basic.StopWatch
  var totalSampledVars = 0
  abstract class IterState {
    var temperature = 1.0
    var numSampledVars = 0
    var numChangedVars = 0
    val acceptProbFig = new fig.basic.BigStatFig
    val blockSizeFig = new fig.basic.BigStatFig

    def log = {
      putLogRec("currIter", iter) // Time measured in number of iterations
      putLogRec("currTime", timer.getCurrTimeLong/1000) // Time measured in seconds
      putLogRec("temperature", temperature)

      putLogRec("totalSampledVars", totalSampledVars) // Time measured in number of variables sampled
      putLogRec("numSampledVars", numSampledVars) // Number of variables sampled this iteration
      putLogRec("numChangedVars", numChangedVars) // Number that actually changed value
      putLogRec("changedVarsFrac", 1.0*numChangedVars/numSampledVars) // Average number of times a variable was sampled
      putLogRec("acceptProb", acceptProbFig) // If MH is used, acceptance probabilities
      putLogRec("blockSize", blockSizeFig) // If MH is used, acceptance probabilities
    }

    // Return whether to continue running
    def run : Boolean = {
      Utils.track_foreach("Example", trainExamples, { ex:Example =>
        if (opts.verbose >= 6) Utils.track_printAll("State") { ex.log }
        processExample(ex)
        if (opts.repCheck >= 3) track_repCheck
        true
      })
      totalSampledVars += numSampledVars
      if (burnedIn_?(iter)) updateInferenceStats
      sampleHyperparameters
      true
    }

    def processExample(ex:Example) : Unit
    def updateInferenceStats : Unit
    def sampleHyperparameters = { }

    def annealProb(w1:Double, w2:Double) = {
      if (temperature == 1) w1/(w1+w2)
      else {
        val u1 = math.pow(w1, 1.0/temperature)
        val u2 = math.pow(w2, 1.0/temperature)
        u1/(u1+u2)
      }
    }
  }
  def newIterState : IterState

  // Assume source and dest are file names in the execution directory
  def createSymLink(source:String, dest:String) = {
    fig.basic.Utils.systemHard("ln -sf "+source+" "+Execution.getFile(dest))
  }

  def learn = {
    def writeStuff(writeNow:Boolean) = {
      if (!isIntermediateLearner && (writeNow || iter % opts.writeInferenceStatsInterval == 0)) {
        Utils.writeLines(Execution.getFile("params."+iter), { puts => params.write(opts.paramsMaxPrint, puts) })
        Utils.writeLines(Execution.getFile("avgParams."+iter), { puts => avgParams.write(opts.paramsMaxPrint, puts) })
        if (opts.writeInferenceStats) writeInferenceStats(Execution.getFile("inferenceStats."+iter))
        if (opts.writePredictions) Utils.writeLines(Execution.getFile("pred."+iter), { puts:(String=>Any) =>
          (0 to trainExamples.size-1).foreach { i => puts("=== Example "+i); trainExamples(i).write(puts) }
        })

        // Link to the latest (so we can easily access the current state before the run finishes for evaluation, inspection etc.)
        createSymLink("avgParams."+iter, "avgParams.final")
        createSymLink("params."+iter, "params.final")
        if (opts.writeInferenceStats) createSymLink("inferenceStats."+iter, "inferenceStats.final")
        if (opts.writePredictions) createSymLink("pred."+iter, "pred.final")
     
        // Write grammar
        if (opts.evalParsing) {
          track("Writing grammar") { Utils.writeLines(parsingGrammarPrefix+"."+iter, avgParams.writeGrammar _) }
          track("Writing lexicon") { lexicon.write(parsingLexiconPrefix+"."+iter) }
          createSymLink(parsingGrammarNamePrefix+"."+iter, parsingGrammarNamePrefix+".final")
          createSymLink(parsingLexiconNamePrefix+"."+iter, parsingLexiconNamePrefix+".final")
        }
      }
      log
    }

    track("learn: Initialize") {
      track("Run") { learnInit }
      learnPostInit
      Record.begin("iteration", 0)
      Utils.track_printAll("Stats") { writeStuff(true) }
      Record.end
    }

    val stateMap = new CounterMap[String] // For small examples, can store this

    track("learn: Iterating") {
      timer.start
      // The last iteration is just for printing out stuff, not training
      var inprogress = true
      while (inprogress && iter < opts.maxIters) {
        iter += 1
        begin_track("Iteration %s/%s", iter, opts.maxIters)
        Record.begin("iteration", iter)

        // Create state
        val iterState = newIterState
        iterState.temperature = {
          if (opts.maxIters == 1 || burnedIn_?(iter))
            opts.finalTemperature
          else
            opts.initTemperature + (opts.finalTemperature-opts.initTemperature) * iter / (opts.burnInFrac*opts.maxIters - 1)
        }

        // Do learning
        Utils.track("Run") { inprogress = iterState.run }

        if (opts.accumulateFullStats)
          stateMap.incr(trainExamples.map(stateStr(_)).mkString(" | "), 1)

        Utils.track_printAll("Stats") {
          iterState.log
          writeStuff(!inprogress)
        }
        if (opts.repCheck >= 2) track_repCheck

        Record.end
        end_track
      }
    }

    if (opts.accumulateFullStats) Utils.track_printAll("State stats") {
      stateMap.write(Integer.MAX_VALUE, true, "", logs("%s", _))
    }

    learnFinish
  }

  def learnInit = { }
  def learnPostInit = {
    if (opts.repCheck >= 1) track_repCheck
    if (trainExamples != null) putLogRec("train.numExamples", trainExamples.size)
    if (testExamples != null) putLogRec("test.numExamples", testExamples.size)
  }

  def learnFinish = {
    if (opts.verbose >= 6) Utils.track_printAll("Final states") {
      Utils.track_foreach("Example", trainExamples, { ex:Example =>
        Utils.track_printAll("State") { ex.log }
        true
      })
    }

    // Evaluate parsing
    if (!isIntermediateLearner && opts.evalParsing) {
      fig.exec.Execution.runSystemCommand("Evaluating parsing performance",
        "eval-parsing/parse "+parsingGrammarPrefix+".final "+opts.evalParsingDatasetPrefix)
    }
  }

  def parsingGrammarPrefix = Execution.getFile(parsingGrammarNamePrefix)
  def parsingLexiconPrefix = Execution.getFile(parsingLexiconNamePrefix)
  def parsingGrammarNamePrefix = "avgParams.grammar"
  def parsingLexiconNamePrefix = "avgParams.lexicon"

  def evaluate(tag:String, nodes:Array[Node]) = {
    var sumLogProb = 0.0
    track(tag) {
      track("Examples") {
        Utils.track_foreach("Example", nodes, { node:Node =>
          sumLogProb += logProb(node)
          true
        })
      }
      logs("sumLogProb = %s", sumLogProb)
    }
  }

  def generate(random:Random) : Node = throw new RuntimeException("Not implemented")
  def generate(random:Random, context:Context) : Node = throw new RuntimeException("Not implemented")

  def burnedIn_?(iter:Int) = iter >= opts.burnInFrac*opts.maxIters

  //// CRP stuff

  // Too slow: don't use this
  def crpLogProb(numNew:Int, numMatch:Double, numTotal:Double) = {
    // numNew customers walk into restaurant with numMatch eating the dish they want and numTotal customers already in the table.
    // Return probability of all those customers get the dish they want.
    (logGamma(numMatch + numNew) - logGamma(numMatch)) - (logGamma(numTotal + numNew) - logGamma(numTotal))
  }

  def margLogProb[A](counts:Counter[A], alpha:Double, baseLogProb:A=>Double) = {
    var logProb = 0.0
    val logAlpha = math.log(alpha)
    counts.keys.foreach { a:A =>
      val count = counts.getCount(a)
      if (count > 0) logProb += logAlpha + baseLogProb(a) + logGamma(count)
    }
    logProb += logGamma(alpha) - logGamma(alpha + counts.totalCount)
    logProb
  }

  def sampleAlpha(counts:Counter[_]) = {
    val n = counts.totalCount.toInt
    val k = counts.numNonZero // WARNING: assume number of tables is number of dishes
    val con = new Concentration(opts.alphaPriorRate, n)
    con.doSampleStep(opts.sampleRandom, k)
    logs("sampleAlpha: n=%s, k=%s, alpha = %s", n, k, con.getAlpha)
    con.getAlpha
  }

  def sampleContProb(numCont:Int, numStop:Int) = {
    fig.prob.Beta.sample(opts.sampleRandom, 2*opts.initBaseContProb + numCont, 2*(1-opts.initBaseContProb) + numCont)
  }
}
