package typesampling

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,logs,dbgs}
import java.util.{Set,ArrayList,Random}
import fig.basic.NumUtils.logGamma
import fig.basic.NumUtils.logAdd

object HeuristicSearch {
  def enforceNoDuplicateCalls(n:Int, f:Int=>Double) = {
    val hit = new Array[Boolean](n)
    ({ i:Int => require (!hit(i)); hit(i) = true; f(i) })
  }

  // Too simple - don't use
  // Try to find the optimum value of f(i) for i in 0..n-1
  def argmax(n:Int, f:Int=>Double) : Int = {
    require (n >= 2)
    var left = 0
    var right = n-1
    var left_value = f(left)
    var right_value = f(right)
    while (left+1 < right) {
      val mid = (left+right)/2
      val mid_value = f(mid)
      if (left_value < right_value) {
        left = mid
        left_value = mid_value
      }
      else {
        right = mid
        right_value = mid_value
      }
    }
    if (left_value > right_value) left else right
  }

  // Try to find the optimum value of f(i) for i in 0..n-1
  def iterateArgmax(numIters:Int, n:Int, f:Int=>Double) : Int = {
    require (n >= 2)
    var left = 0
    var right = n-1
    var left_value = f(left)
    var right_value = f(right)
    var far_left = -1
    var far_right = -1
    var far_left_value = Double.NaN
    var far_right_value = Double.NaN
    var iter = 0
    foreach(numIters, { iter:Int =>
      // State: far_left      left      (mid)     right      far_right
      while (left+1 < right) {
        val mid = (left+right)/2
        val mid_value = f(mid)
        if (left_value < right_value) { // Go right 
          far_left = left
          far_left_value = left_value
          left = mid
          left_value = mid_value
        }
        else { // Go left
          far_right = right
          far_right_value = right_value
          right = mid
          right_value = mid_value
        }
      }
      if (iter < numIters-1) { // Continue searching
        if (left_value < right_value) { // Go right
          if (far_right != -1) {
            left = right
            left_value = right_value
            right = far_right
            right_value = far_right_value
            far_left = -1
            far_right = -1
            far_left_value = Double.NaN
            far_right_value = Double.NaN
          }
        }
        else {
          if (far_left != -1) {
            right = left
            right_value = left_value
            left = far_left
            left_value = far_left_value
            far_left = -1
            far_right = -1
            far_left_value = Double.NaN
            far_right_value = Double.NaN
          }
        }
      }
    })
    if (left_value > right_value) left else right
  }

  object EvalNode {
    val invalid = EvalNode(-1, Double.NaN)
  }

  case class EvalNode(i:Int, value:Double) {
    var prev = this
    var next = this
    def hasPrev = prev != this
    def hasNext = next != this
    def insertAfter(node:EvalNode) = {
      if (this.hasNext) {
        node.next = this.next; this.next.prev = node
      }
      node.prev = this; this.next = node
    }
    def argmax : Double = {
      var node = this
      var bestNode = node
      while (node.hasNext) {
        node = node.next
        if (node.value > bestNode.value)
          bestNode = node
      }
      bestNode.i
    }

    override def toString = i+":"+fmt(value) + (if (hasNext) " "+next.toString else "")
  }

  case class ApproxDistrib(logSum:Double, n:Int, head:EvalNode) {
    // The distribution is a mixture of the two:
    // (1-f): the distribution formed by the EvalNode's starting with head
    //     f: the uniform distribution over n
    def sample(random:Random, heldoutFrac:Double) : Int = {
      // Main        Heldout
      // 1*logSum    f/(1-f)*logSum
      // 1 + f/(1-f) = 1/(1-f)
      val logTotalMass = logSum + math.log(1.0/(1-heldoutFrac))
      val logTarget = math.log(random.nextDouble) + logTotalMass
      var logAccum = math.log(0.0)
      var node = head
      while (true) {
        logAccum = logAdd(logAccum, node.value)
        if (logAccum >= logTarget) return node.i
        if (node.hasNext) node = node.next
        else return random.nextInt(n) // Sample from heldout: hopefully shouldn't get here very often
      }
      throw impossible
    }

    def logProb(heldoutFrac:Double, i:Int) = {
      var node = head
      while (node.i < i) { require (node.hasNext); node = node.next } // Find i in the list
      //dbgs("logProb i=%s, n=%s: %s", i, n, if (node.i == i) math.exp(node.value-logSum)*(1-heldoutFrac) else 0.0)
      math.log((if (node.i == i) math.exp(node.value-logSum)*(1-heldoutFrac) else 0.0) + heldoutFrac/n)
    }
  }

  val numEvalsTracker = new InventoryTracker("NumEvals", false)
  val numItersTracker = new InventoryTracker("NumIters", false)
  val numLocalProbesTracker = new InventoryTracker("NumLocalProbes", false)

  def approxDistrib(n:Int, precision:Double, f:Int=>Double) : ApproxDistrib = {
    require (n >= 0)
    var head = EvalNode(0, f(0))
    var logSum = head.value
    var numEvals = 0

    if (n == 1) return ApproxDistrib(logSum, n, head)

    def addNodeAfter(prev:EvalNode, i:Int) = {
      val node = EvalNode(i, f(i))
      numEvals += 1
      logSum = logAdd(logSum, node.value)
      prev.insertAfter(node)
      node
    }

    var left = head
    var right = addNodeAfter(head, n-1)
    var numIters = 0
    while (left.i+1 < right.i) {
      // State: far_left      left      (mid)     right      far_right
      while (left.i+1 < right.i) {
        val mid = addNodeAfter(left, (left.i+right.i)/2)
        if (left.value < right.value) // Go right 
          left = mid
        else // Go left
          right = mid
      }
      numIters += 1

      if (left.value < right.value) { // Go right if possible
        if (right.hasNext) {
          left = right
          right = right.next
        }
      }
      else { // Go left if possible
        if (left.hasPrev) {
          right = left
          left = left.prev
        }
      }
    }

    // Start at the max and fan out
    val top = if (left.value > right.value) left else right
    left = top
    right = top
    var oldLogSum = math.log(0.0)
    var numLocalProbes = 0
    while ((left.hasPrev || right.hasNext) && logSum-oldLogSum >= precision) {
      oldLogSum = logSum
      numLocalProbes += 1
      while (left.hasPrev && left.prev.i == left.i-1) left = left.prev // Skip added nodes
      if (left.hasPrev) {
        addNodeAfter(left.prev, left.i-1)
        left = left.prev
      }
      while (right.hasNext && right.next.i == right.i+1) right = right.next // Skip added nodes
      if (right.hasNext) {
        addNodeAfter(right, right.i+1)
        right = right.next
      }
      //dbgs("EXPAND %s -> %s; left = %s, right = %s", oldLogSum, logSum, left.i, right.i)
    }

    // Record statistics
    if (true) {
      val key = math.pow(10, math.ceil(math.log(n)/math.log(10))).toInt
      numEvalsTracker.add(key, numEvals, null)
      numLocalProbesTracker.add(key, numLocalProbes, null)
      numItersTracker.add(key, numIters, null)
    }

    ApproxDistrib(logSum, n, head)
  }

  def main(args:Array[String]) = {
    // Test
    fig.basic.LogInfo.msPerLine = 0
    val n = 40
    val distrib = approxDistrib(n, 1e-4, { i:Int =>
      val v = -(i-3).abs // Maximum at 3
      //val v = 0 // Flat
      //val v = (i-n/2) // Minima at n/2 (only end up getting one side)
      //val v = -i // Mode at 0
      logs("f(%s) = %s", i, v)
      v
    })
    val counts = new CounterMap[Any]
    val rand = new Random(1)
    val heldoutFrac = 0
    foreach(100000, { i:Int =>
      counts.incr(distrib.sample(rand, heldoutFrac), 1)
    })
    foreach(n, { i:Int =>
      logs("logProb(%s) = %s", i, distrib.logProb(heldoutFrac, i))
    })
    counts.write(Integer.MAX_VALUE, false, "", logs("%s", _))
  }
}

object Misc {
  def indentStr(n:Int) : String = if (n == 0) "" else "  "+indentStr(n-1)

  // Return a list with removed from the given list
  def removeFirst[A](list:List[A], x:A) : List[A] = {
    list match {
      case Nil => Nil
      case y :: ys => if (x == y) ys else y :: removeFirst(ys, x)
    }
  }
  def addToList[A,B](map:HashMap[A,List[B]], key:A, value:B) = {
    map(key) = value :: map.getOrElse(key, Nil)
  }

  // Don't return -infty for log(0)
  def smoothLog(x:Double) = {
    if (x > 0) math.log(x)
    else -10000.0
  }

  def I[T](x:T,y:T) = (if (x == y) 1 else 0)
  def logChoose(n:Int, k:Int) = logGamma(n+1) - logGamma(k+1) - logGamma(n-k+1)
  def logGenFact(a:Double, n:Int) : Double = {
    if (n == 0) { require (a >= 0, a); 0.0 }
    else {
      require (a > 0 && n >= 0, a + " "+ n)
      logGamma(a+n) - logGamma(a)
    }
  }
  def logSumExp(xs:Array[Double]) = {
    val offset = Utils.max(xs)
    offset + math.log(Utils.sum(xs.map { x => math.exp(x-offset) }))
  }

  def fmt1_compress(threshold:Double, probs:Array[Double]) = {
    def recurse(i:Int, streak:Int, output:List[String]) : List[String] = {
      if (i == probs.size) {
        if (streak > 0) "..["+streak+"].." :: output else output
      }
      else {
        if (probs(i) < threshold)
          recurse(i+1, streak+1, output)
        else
          recurse(i+1, 0, fmt(probs(i)) :: (if (streak > 0) "..["+streak+"].." :: output else output))
      }
    }
    recurse(0, 0, Nil).reverse.mkString(" ")
  }
}

class Accumulator {
  private var sumValue = 0.0
  private var n = 0
  def value = sumValue / n
  def +=(dv:Double) = { sumValue += dv; n += 1 }
  override def toString = fmt(value)
}

// Map a key to a list of items (assume distinct)
class ListHashMap[A,B <: Any] {
  val map = new HashMap[A, List[B]]
  def add(key:A, values:List[B]) = {
    if (values != Nil) {
      map.get(key) match {
        case None => map(key) = values
        case Some(oldValues) => map(key) = values ++ oldValues
      }
    }
  }
  def remove(key:A, matches:B=>Boolean) = {
    map.get(key) match {
      case None => Nil
      case Some(values) =>
        val (l1, l2) = values.partition(matches) // l1: removed and returned; l2: not removed and kept
        // Keep ones that don't match
        if (l2 == Nil) map -= key 
        else map(key) = l2
        l1 // Return ones that match
    }
  }

  def size = map.size

  def add(key:A, value:B) = Misc.addToList(map, key, value)
  // Assume exactly one value matches
  /*def removeIdentity(key:A, value:B) = {
    val oldValues = map.getOrElse(key, Nil)
    val newValues = Misc.removeFirstIdentity(oldValues, value)
    if (newValues == Nil) map -= key
    else map(key) = newValues
    require (oldValues.size == newValues.size+1, oldValues.size + " " + newValues.size + " " + key + " " + value)
  }*/

  // Assume exactly one value matches
  def remove(key:A, value:B) = {
    val oldValues = map.getOrElse(key, Nil)
    val newValues = Misc.removeFirst(oldValues, value)
    if (newValues == Nil) map -= key
    else map(key) = newValues
    require (oldValues.size == newValues.size+1, oldValues.size + " " + newValues.size + " " + key + " " + value)
  }

  def getCount(a:A) = {
    map.get(a) match {
      case None => 0
      case Some(l) => l.size
    }
  }
}

object BigDoubleArray {
  val BASE = 1e10
  val LOG_BASE = math.log(BASE)
  val INV_BASE = 1.0/BASE
  val BASE_POW = map(30, { i:Int => math.pow(BASE, i) })
}
class BigDoubleArray(N:Int) {
  import BigDoubleArray._
  val weights = map(N, 1.0)
  val exps = map(N, {_:Int => 0})

  // For each i = 0..N-1, entry i is base * ... * (base+i*d-1)
  def multGenFact(sign:Int, base:Double, d:Int) = {
    require (base > 0 && !base.isInfinite, base)
    var M = 1.0
    var E = 0
    foreach(N, { i:Int =>
      if (sign == +1) {
        weights(i) *= M
        exps(i) += E
      }
      else if (sign == -1) {
        weights(i) /= M
        exps(i) -= E
      }
      else
        throw impossible
      flush(i)
      if (d == 1) M *= base + i
      else if (d == 2) M *= (base + 2*i)*(base + 2*i+1)
      else throw impossible
      while (M > BASE) { M /= BASE; E += 1 }
      while (M < INV_BASE) { M *= BASE; E -= 1 }
    })
  }

  // For each i = 0..N-1, entry N-i-1 is base * ... * (base+i*d-1)
  def multGenFactReverse(sign:Int, base:Double, d:Int) = {
    require (base > 0 && !base.isInfinite, base)
    var M = 1.0
    var E = 0
    //dbgs("%s %s %s", sign, base, d)
    foreach(N, { i:Int =>
      if (sign == +1) {
        //dbgs("%s M=%s", weights(N-i-1), M)
        weights(N-i-1) *= M
        exps(N-i-1) += E
      }
      else if (sign == -1) {
        weights(N-i-1) /= M
        exps(N-i-1) -= E
      }
      else
        throw impossible
      flush(N-i-1)
      if (d == 1) M *= base + i
      else if (d == 2) M *= (base + 2*i)*(base + 2*i+1)
      else throw impossible
      while (M > BASE) { M /= BASE; E += 1 }
      while (M < INV_BASE) { M *= BASE; E -= 1 }
    })
  }

  private def flush(i:Int) = {
    Utils.assertValid(weights(i))
    while (weights(i) > BASE) { weights(i) /= BASE; exps(i) += 1 }
    if (weights(i) > 0) {
      while (weights(i) < INV_BASE) { weights(i) *= BASE; exps(i) -= 1 }
    }
  }

  def multExp(score:Double) = {
    val E = (score / LOG_BASE).toInt
    val M = math.exp(score-E*LOG_BASE)
    foreach(N, { i:Int =>
      weights(i) *= M
      exps(i) += E
    })
  }

  def pow(x:Double) = {
    throw Utils.TODO
  }

  def normalize : Array[Double] = {
    //foreach(N, { i:Int => flush(i) })
    val maxE = Utils.max(exps)
    foreach(N, { i:Int =>
      val E = exps(i) - maxE
      if (E > 0) weights(i) *= BASE_POW(E)
      else if (E < 0) {
        if(-E < BASE_POW.length) weights(i) /= BASE_POW(-E)
        else weights(i) = 0 // If too small, just ignore
      }
      exps(i) = 0
    })
    Utils.normalize_!(weights)
    weights
  }

  def toLog = map(N, { i:Int => exps(i) * LOG_BASE + math.log(weights(i)) })
}

// Keep track and print out the maximum value and sum
class InventoryTracker(name:String, reverse:Boolean) {
  val bestMap = new HashMap[Any,Double]
  val countMap = new CounterMap[Any]
  val meanMap = new CounterMap[Any]
  var best_v = Double.NaN

  def cmpStr = if (reverse) "min" else "max"
  def better(v1:Double, v2:Double) = if (reverse) v1 < v2 else v1 > v2
  
  def add(a:Any) : Boolean = add(a, 0, null)

  def add(a:Any, v:Double, details:Any) : Boolean = {
    countMap.incr(a, 1)
    meanMap.incr(a, v)
    if (best_v.isNaN || better(v, best_v)) best_v = v
    val updated = bestMap.get(a) match {
      case None => bestMap(a) = v; true
      case Some(old_v) =>
        if (better(v, old_v)) { bestMap(a) = v; true }
        else false
    }
    if (updated)
      fig.basic.LogInfo.rant(fmts("INVENTORY %s: (global)[%s=%s, mean=%s, count=%s]; %s[%s=%s, mean=%s, count=%s]%s",
        name, cmpStr, best_v, meanMap.totalCount/countMap.totalCount, countMap.totalCount,
        a, cmpStr, v, meanMap.getCount(a)/countMap.getCount(a), countMap.getCount(a),
        if (details == null) "" else "; "+details))
    updated
  }

  def log = Utils.track_printAll("Inventory %s: %s=%s, mean=%s, count=%s",
      name, cmpStr, best_v, meanMap.totalCount/countMap.totalCount, countMap.totalCount) {
    bestMap.keys.toList.sortWith { (a,b) => better(bestMap(a), bestMap(b)) }.foreach { a:Any =>
      Utils.logs("%s: %s=%s, mean=%s, count=%s", a, cmpStr, bestMap(a), meanMap.getCount(a)/countMap.getCount(a), countMap.getCount(a))
    }
  }
}
