package typesampling.ptsg

import typesampling._
import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,fail,impossible}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import fig.basic.NumUtils.logGamma
import Misc.{I,logChoose,logGenFact}

// KNode: an elementary tree (used for a key)
trait KNode {
  val context : String
  def foreachNode(f:KNode=>Any) : Unit
}
case class KTruncatedNode(context:String) extends KNode {
  def foreachNode(f:KNode=>Any) = f(this)
  override def toString = context+"|"
}
case class KPrimitiveNode(context:String, value:AnyRef) extends KNode {
  def foreachNode(f:KNode=>Any) = f(this)
  override def toString = "{"+context+" "+value+"}"
}
case class KCompositeNode(context:String, left:KNode, right:KNode) extends KNode {
  def foreachNode(f:KNode=>Any) = { f(this); left.foreachNode(f); right.foreachNode(f) }
  override def toString = "{"+context+" "+left+" "+right+"}"
}

////////////////////////////////////////////////////////////

trait CutNode extends AbstractNode[String,CutNode] {
  var cut = true // Whether this node is the root of a subtree fragment
  // Store sampling statistics
  var cutCount = 0
  var uncutCount = 0
  var lastIterSampled = -1

  var parentOption : Option[CutNode] = None
  def parent = parentOption.getOrElse(throw impossible)
  // Lowest common ancestor that's cut (not including self)
  def cutLCA : CutNode = cutLCA(true)
  def cutLCA(first:Boolean) : CutNode = {
    if (first || !cut) parent.cutLCA(false)
    else this
  }

  def cutProb = 1.0*cutCount/(cutCount+uncutCount)

  def updateCounts = {
    if (cut) cutCount += 1
    else uncutCount += 1
  }

  // Functions for navigating the tree.
  // Directions can be followed if we do not go through any cut points.
  def matches(query:KNode) : Boolean = matches(query, true)
  def matches(node:KNode, first:Boolean) : Boolean // if truncatedTree == node (but cut doesn't have to be set), but do this more efficiently than that
  def getDirectionsTo(dest:CutNode) : Option[List[Boolean]] = getDirectionsTo(dest, true)
  def getDirectionsTo(dest:CutNode, first:Boolean) : Option[List[Boolean]] // find directions leading to something that matches node (use cut to prune)
  def followDirections(directions:List[Boolean]) : Option[CutNode] = followDirections(directions, true)
  def followDirections(directions:List[Boolean], first:Boolean) : Option[CutNode]

  // Return a copy of the node
  def truncatedTree : KNode // Return a tree whose leaves have cut = true (don't truncate the root)
  def immediateTruncatedTree : KNode // Return a tree whose leaves have cut = true (helper function)

  // For each node and the root node of its subtree (lowest ancestor for which cut = true, not including this)
  /*def foreachNodeAndSubRootNode(f:(CutNode,CutNode)=>Any) : Unit = {
    require (cut)
    foreachNodeAndSubRootNode(f, this)
  }
  def foreachNodeAndSubRootNode(f:(CutNode,CutNode)=>Any, subRootNode:CutNode) : Unit // Helper function*/

  def wrapCut(s:String) = wrapCutLeft + s + wrapCutRight
  def wrapCutLeft = fmt(cutProb) + (if (cut) '[' else '(')
  def wrapCutRight = (if (cut) ']' else ')')
}

class CutPrimitiveNode(_context:String, _value:AnyRef) extends CutNode with AbstractPrimitiveNode[String,CutNode] {
  val context = _context
  val value = _value
  def truncatedTree = { // first
    require (cut)
    KPrimitiveNode(context, value)
  }
  def immediateTruncatedTree = { // not first
    if (cut) KTruncatedNode(context)
    else KPrimitiveNode(context, value)
  }
  //def foreachNodeAndSubRootNode(f:(CutNode,CutNode)=>Any, subRootNode:CutNode) = f(this, subRootNode)

  def matches(query:KNode, first:Boolean) = {
    query match {
      case query:KTruncatedNode => cut && context == query.context
      case query:KPrimitiveNode => (first || !cut) && context == query.context && value == query.value
      case query:KCompositeNode => false
    }
  }
  def getDirectionsTo(dest:CutNode, first:Boolean) = {
    if (this eq dest) Some(Nil)
    else None
  }
  def followDirections(directions:List[Boolean], first:Boolean) = {
    directions match {
      case Nil => Some(this) // Stop here
      case _ => None
    }
  }

  override def toString = wrapCut(context+" "+value)
}

class CutCompositeNode(_context:String, _left:CutNode, _right:CutNode) extends CutNode with AbstractCompositeNode[String,CutNode] {
  val context = _context
  val left = _left
  val right = _right
  _left.parentOption = Some(this)
  _right.parentOption = Some(this)
  def truncatedTree = { // first
    require (cut)
    KCompositeNode(context, left.immediateTruncatedTree, right.immediateTruncatedTree)
  }
  def immediateTruncatedTree = { // not first
    if (cut) KTruncatedNode(context)
    else KCompositeNode(context, left.immediateTruncatedTree, right.immediateTruncatedTree)
  }

  def matches(query:KNode, first:Boolean) = {
    query match {
      case query:KTruncatedNode => cut && context == query.context
      case query:KPrimitiveNode => false
      case query:KCompositeNode => (first || !cut) && context == query.context && left.matches(query.left, false) && right.matches(query.right, false)
    }
  }
  def getDirectionsTo(node:CutNode, first:Boolean) = {
    if (this eq node) Some(Nil)
    else if (first || !cut) {
      left.getDirectionsTo(node, false) match {
        case Some(directions) => Some(false :: directions)
        case None =>
          right.getDirectionsTo(node, false) match {
            case Some(directions) => Some(true :: directions)
            case None => None
          }
      }
    }
    else None
  }
  def followDirections(directions:List[Boolean], first:Boolean) = {
    directions match {
      case Nil => Some(this)
      case dir :: restDirections =>
        if (first || !cut) {
          if (!dir) left.followDirections(restDirections, false)
          else right.followDirections(restDirections, false)
        }
        else
          None
    }
  }

  override def toString = wrapCut(context+" "+left+" "+right)
  override def write(puts:String=>Any, indent:Int) = {
    puts(Misc.indentStr(indent) + wrapCutLeft+context)
    left.write(puts, indent+1)
    right.write(puts, indent+1)
    puts(Misc.indentStr(indent) + wrapCutRight)
  }
}

////////////////////////////////////////////////////////////

case class PTSGLearner(opts:LearnOptions) extends Learner {
  type Example = CutNode

  def convertNode(node:Node) : CutNode = {
    node match {
      case node:PrimitiveNode => new CutPrimitiveNode(pcfgLearner.convertContext(node.context), node.value)
      case node:CompositeNode => new CutCompositeNode(pcfgLearner.convertContext(node.context), convertNode(node.left), convertNode(node.right))
    }
  }

  override def setTrainNodes(trainNodes:Array[Node]) = {
    super.setTrainNodes(trainNodes) // Calls convertNode
    pcfgLearner.setTrainNodes(trainNodes)
  }

  val pcfgLearner = new pcfg.StringPCFGLearner(opts) // For the base distribution
  pcfgLearner.isIntermediateLearner = true

  class Params extends AbstractParams {
    val base = pcfgLearner.params
    val alphas = new HashMap[String,Double]
    val baseContProbs = new HashMap[String,Double]
    val treeCounts = new CounterMapCollection[String,KNode]

    def write(maxPrint:Int, puts:String=>Any) = {
      base.write(maxPrint, puts)
      alphas.keys.foreach { context:String =>
        puts("ALPHA\t"+context+"\t"+alphas(context))
      }
      baseContProbs.keys.foreach { context:String =>
        puts("CONT\t"+context+"\t"+baseContProbs(context))
      }
      treeCounts.write(maxPrint, maxPrint, true, "TREE\t", puts)
    }

    // Write grammar for evaluating parsing using the Berkeley Decoder
    override def writeGrammar(puts:String=>Any) = {
      base.startCounts.keys.foreach { context:String =>
        puts(fmts("(ROOT %s[0]) :: %s[0] ||| logProb:%s count:1", context, context, math.log(base.startCounts.getProb(context))))
      }
      treeCounts.values.foreach { map:Counter[KNode] =>
        map.keys.foreach { t:KNode =>
          var index = 0
          val leaves = new ArrayBuffer[String]
          def renderTree(t:KNode) : String = {
            t match {
              case t:KTruncatedNode =>
                val str = t.context+"["+index+"]"
                leaves += str
                index += 1
                str
              case t:KPrimitiveNode =>
                val str = "("+t.context+" "+t.value+")"
                leaves += t.value.toString
                str
              case t:KCompositeNode =>
                val str = "("+t.context+" "+renderTree(t.left)+" "+renderTree(t.right)+")"
                str
            }
          }
          val str = renderTree(t)
          puts(fmts("%s :: %s ||| logProb:%s count:1", str, leaves.mkString(" "), math.log(map.getProb(t))))
        }
      }
    }

    def size = treeCounts.numNonZero

    // this = avgParams, that = params (current)
    // params contains a set of counts; together with alpha, this defines a predictive distribution over trees (for each context)
    // We add this distribution to avgParams
    def add(that:Params) = track("Adding in predictive distribution") {
      // Update avgParams
      that.treeCounts.keys.foreach { context:String => // For each context...
        val map = that.treeCounts(context) // These are the counts
        val alpha = that.alphas(context)
        val z = map.totalCount + alpha // Normalization constant for the predictive distribution
        logs("Normalization: %s", z)

        // Add the counts
        this.treeCounts.incr(context, 1.0/z, map)

        // Now add things from base distribution
        // For now, just add simple PCFG rules
        that.base.lexCounts.get(context) match {
          case Some(lexMap) =>
            val lexProb = that.base.choiceCounts.getProb(context, pcfgLearner.C_LEX)
            lexMap.keys.foreach { value =>
              val prob = lexProb * lexMap.getProb(value)
              val node = KPrimitiveNode(context, value.asInstanceOf[String])
              this.treeCounts.incr(context, node, alpha/z * prob)
            }
          case None =>
        }
        that.base.ruleCounts.get(context) match {
          case Some(ruleMap) =>
            val ruleProb = that.base.choiceCounts.getProb(context, pcfgLearner.C_RULE)
            ruleMap.keys.foreach { ruleOutput =>
              val prob = ruleProb * ruleMap.getProb(ruleOutput)
              val node = KCompositeNode(context, KTruncatedNode(ruleOutput.left), KTruncatedNode(ruleOutput.right))
              this.treeCounts.incr(context, node, alpha/z * prob)
            }
          case None =>
        }
      }
    }
  }

  val params = new Params
  val avgParams = new Params
  val treeLocs = new ListHashMap[KNode,CutNode] // type -> tokens

  // Probability of generating this tree from the base distribution
  // At each node (except the root), we choose whether to stop or not.
  def baseProb(node:KNode) : Double = baseProb(node, true)
  def baseProb(node:KNode, mustCont:Boolean) : Double = {
    val contProb = { if (mustCont) 1.0 else params.baseContProbs(node.context) }
    node match {
      case node:KTruncatedNode => require (!mustCont); 1-contProb
      case node:KPrimitiveNode =>
        contProb *
        params.base.choiceCounts.getProb(node.context, pcfgLearner.C_LEX) *
        params.base.lexCounts.getProb(node.context, node.value)
      case node:KCompositeNode =>
        contProb *
        params.base.choiceCounts.getProb(node.context, pcfgLearner.C_RULE) *
        params.base.ruleCounts.getProb(node.context, pcfg.RuleOutput(node.left.context, node.right.context)) *
        baseProb(node.left, false) * baseProb(node.right, false)
    }
  }
  def baseLogProb(node:KNode) : Double = baseLogProb(node, true)
  def baseLogProb(node:KNode, mustCont:Boolean) : Double = {
    val contProb = { if (mustCont) 1.0 else params.baseContProbs(node.context) }
    node match {
      case node:KTruncatedNode => require (!mustCont); 1-contProb
      case node:KPrimitiveNode =>
        math.log(contProb *
        params.base.choiceCounts.getProb(node.context, pcfgLearner.C_LEX) *
        params.base.lexCounts.getProb(node.context, node.value))
      case node:KCompositeNode =>
        math.log(contProb *
        params.base.choiceCounts.getProb(node.context, pcfgLearner.C_RULE) *
        params.base.ruleCounts.getProb(node.context, pcfg.RuleOutput(node.left.context, node.right.context))) +
        baseLogProb(node.left, false) + baseLogProb(node.right, false)
    }
  }

  def repCheck = track("repCheck") {
    def change(dv:Double) = {
      var numPieces = 0
      trainExamples.foreach { rootNode:CutNode =>
        rootNode.foreachNode { node:CutNode =>
          if (node.cut) {
            val t = node.truncatedTree
            params.treeCounts.incr(t.context, t, dv)
            numPieces += 1
          }
        }
      }
      numPieces
    }
    val numPieces = change(-1)
    require (params.treeCounts.numKeys == 0)
    change(+1)

    // Check that treeLocs is consistent with treeCounts
    if (opts.typeLevelSampling) {
      //require (treeLocs.map.size == params.treeCounts.size, treeLocs + " " + params.treeCounts)
      var numTrees = 0
      params.treeCounts.keys.foreach { context:String =>
        val map = params.treeCounts(context)
        
        map.keys.foreach { tree:KNode =>
          numTrees += 1
          val locs = treeLocs.map(tree)
          require (locs.size == map.getCount(tree))
          treeLocs.map(tree).foreach { realNode:CutNode =>
            require (realNode.truncatedTree == tree)
          }
        }
      }
      require (numTrees == treeLocs.size, numTrees + " " + treeLocs.size)
    }

    logs("repCheck passed: numPieces = %s", numPieces)
  }

  def logState = track("State") {
    trainExamples.foreach { node:CutNode =>
      logs("%s", node)
    }
    treeLocs.map.keys.foreach { tree:KNode =>
      track("%s", tree) {
        treeLocs.map(tree).foreach { node:CutNode => logs("%s", node) }
      }
    }
  }
  
  def stateLogProb : Double = {
    var logProb = 0.0
    params.treeCounts.keys.foreach { context:String => // For each context (restaurant)
      val map = params.treeCounts(context)
      val alpha = params.alphas(context)
      map.keys.foreach { node:KNode =>
        logProb += math.log(alpha) + baseLogProb(node)
        logProb += logGamma(map.getCount(node))
      }
      logProb += logGamma(alpha) - logGamma(alpha + map.totalCount)
    }
    logProb
  }

  def getInferenceStats(node:CutNode) = {
    val stats = new ArrayBuffer[Double]
    node.foreachNode { node:CutNode =>
      stats += node.cutProb
    }
    stats.toArray
  }

  ////////////////////////////////////////////////////////////
  // Learning

  override def learnInit = {
    super.learnInit
    def initNode(rootNode:CutNode) = {
      opts.initType match {
        case Enums.InitType.random =>
          rootNode.foreachNode(_.cut = opts.initRandom.nextDouble < opts.initCutProb)
          rootNode.cut = true
      }
    }
    track("Base distribution") {
      // When continue, generate according to PCFG probabilities
      pcfgLearner.learn

      // Initialize contination probabilities and alpha
      def f(c:String) = {
        params.baseContProbs(c) = opts.initBaseContProb
        params.alphas(c) = opts.alpha
      }
      params.base.lexCounts.keys.foreach(f(_))
      params.base.ruleCounts.keys.foreach(f(_))
    }

    // Initialize trees
    track("Initializing trees") {
      Utils.track_foreach("Example", trainExamples, { rootNode:CutNode =>
        initNode(rootNode)
        rootNode.foreachNode { node:CutNode =>
          if (node.cut) {
            params.treeCounts.incr(node.context, node.truncatedTree, 1)
            if (opts.typeLevelSampling)
              treeLocs.add(node.truncatedTree, node)
          }
        }
        true
      })
    }
    avgParams.add(params) // TMP
  }

  override def learnFinish = {
    super.learnFinish
    Utils.writeLines(Execution.getFile("avgParams.grammar"), avgParams.writeGrammar _)
  }

  class IterState extends super.IterState {
    def setCut(node:CutNode, newCut:Boolean) = {
      node.lastIterSampled = iter
      numSampledVars += 1
      if (node.cut != newCut) {
        node.cut = newCut
        numChangedVars += 1
      }
    }

    override def sampleHyperparameters = {
      if (opts.sampleAlpha) track("sampleAlpha") {
        params.alphas.keys.foreach { context:String =>
          params.alphas(context) = sampleAlpha(params.treeCounts.get(context) match {
            case Some(counts) => counts
            case None => new CounterMap[KNode]
          })
        }
      }
      if (opts.sampleBaseContProb) track("sampleBaseContProb") {
        params.baseContProbs.keys.foreach { context:String =>
          var n0 = 2 * (1-opts.initBaseContProb)
          var n1 = 2 * (opts.initBaseContProb)
          // For each tree, count number of continues and stops
          params.treeCounts.get(context) match {
            case Some(counts) =>
              counts.keys.foreach { rootNode:KNode =>
                rootNode.foreachNode { node:KNode =>
                  node match {
                    case node:KTruncatedNode => n0 += 1
                    case _ => n1 += 1
                  }
                }
                n1 -= 1 // Don't count the root
              }
            case None =>
          }
          params.baseContProbs(context) = fig.prob.Beta.sample(opts.sampleRandom, n1, n0)
        }
      }
    }

    def processExample(rootNode:CutNode) = {
      rootNode.foreachNode { node:CutNode =>
        if (node != rootNode) {
          if (!opts.sampleSiteOncePerIter || node.lastIterSampled != iter)
            sampleCut(node, node.cutLCA, burnedIn_?(iter))
        }
      }
    }
    override def updateInferenceStats = {
      trainExamples.foreach { rootNode:CutNode =>
        rootNode.foreachNode(_.updateCounts)
      }

      avgParams.add(params)
    }

    def sampleCut(node:CutNode, subRootNode:CutNode, burnedIn:Boolean) = {
      if (opts.typeLevelSampling) sampleCutTypeLevel(node, subRootNode, burnedIn)
      else sampleCutTokenLevel(node, subRootNode, burnedIn)
    }

    def sampleCutTokenLevel(node:CutNode, subRootNode:CutNode, burnedIn:Boolean) = {
      // Consider cutting or uncutting this node
      // If cut, then get two trees (t1, t2)
      // If uncut, then get one tree (t3)
      val (t1, t2, t3) = {
        if (node.cut) {
          val t1 = subRootNode.truncatedTree 
          val t2 = node.truncatedTree
          params.treeCounts.incr(t1.context, t1, -1)
          params.treeCounts.incr(t2.context, t2, -1)
          if (opts.typeLevelSampling) {
            treeLocs.remove(t1, subRootNode)
            treeLocs.remove(t2, node)
          }
          node.cut = false
          val t3 = subRootNode.truncatedTree
          node.cut = true
          (t1, t2, t3)
        }
        else {
          val t3 = subRootNode.truncatedTree
          params.treeCounts.incr(t3.context, t3, -1)
          if (opts.typeLevelSampling)
            treeLocs.remove(t3, subRootNode)
          node.cut = true
          val t1 = subRootNode.truncatedTree 
          val t2 = node.truncatedTree
          node.cut = false
          (t1, t2, t3)
        }
      }

      val c1 = t1.context
      val c2 = t2.context
      val c3 = t3.context

      val alpha1 = params.alphas(c1)
      val alpha2 = params.alphas(c2)
      val alpha3 = params.alphas(c3)
      val p1 = baseProb(t1)
      val p2 = baseProb(t2)
      val p3 = baseProb(t3)
      val nt1 = params.treeCounts.getCount(c1, t1) + alpha1*p1
      val nt2 = params.treeCounts.getCount(c2, t2) + alpha2*p2
      val nt3 = params.treeCounts.getCount(c3, t3) + alpha3*p3
      val nc1 = params.treeCounts.totalCount(c1) + alpha1
      val nc2 = params.treeCounts.totalCount(c2) + alpha2
      val nc3 = params.treeCounts.totalCount(c3) + alpha3

      val cut1Weight = nt1/nc1 
      val cut2Weight = (nt2 + I(t1, t2))/(nc2 + I(c1, c2))
      val cutWeight = cut1Weight * cut2Weight
      val uncutWeight = nt3/nc3
      val cutProb = cutWeight / (cutWeight + uncutWeight)
      val uncutProb = 1 - cutProb

      setCut(node, opts.sampleRandom.nextDouble < cutProb)
      if (node.cut) {
        params.treeCounts.incr(c1, t1, 1)
        params.treeCounts.incr(c2, t2, 1)
        if (opts.typeLevelSampling) {
          treeLocs.add(t1, subRootNode)
          treeLocs.add(t2, node)
        }
      }
      else {
        params.treeCounts.incr(c3, t3, 1)
        if (opts.typeLevelSampling)
          treeLocs.add(t3, subRootNode)
      }

      if (opts.verbose >= 5) track("SAMPLE") {
        logs("CUT %s * %s = %s (prob %s): %s AND %s", cut1Weight, cut2Weight, cutWeight, cutProb, t1, t2)
        logs("UNCUT %s (prob %s): %s", uncutWeight, uncutProb, t3)
        logs("PROBS: %s %s", uncutProb, cutProb)
        //def aa(ex:Example) = ex match { case node:CutCompositeNode => node.left.cut + " " + node.right.cut; case _ => throw impossible }
        //logs(aa(trainExamples(0)) + " " + aa(trainExamples(1)) + " HAS PROBS: %s %s", uncutProb, cutProb)
        logs("CHOOSE cut = %s", node.cut)
        //logState
      }
    }

    def sampleCutTypeLevel(node:CutNode, subRootNode:CutNode, burnedIn:Boolean) = {
      val (t1, t2, t3) = {
        if (node.cut) {
          val t1 = subRootNode.truncatedTree 
          val t2 = node.truncatedTree
          node.cut = false
          val t3 = subRootNode.truncatedTree
          node.cut = true
          (t1, t2, t3)
        }
        else {
          val t3 = subRootNode.truncatedTree
          node.cut = true
          val t1 = subRootNode.truncatedTree 
          val t2 = node.truncatedTree
          node.cut = false
          (t1, t2, t3)
        }
      }
      if (t1 == t2) {
        logs("Got %s, deferring to token level sampling", t3)
        sampleCutTokenLevel(node, subRootNode, burnedIn)
      }
      else {
        //if (opts.verbose >= 5) logState
        //val oldStateStr = tmpStateStr

        // Three things to update: treeLocs, treeCounts, ex.cut

        // How to get from t1 to t2
        val directions = subRootNode.getDirectionsTo(node).getOrElse(throw fail("Can't get to " + node))
        def follow(node:CutNode) = node.followDirections(directions).getOrElse(throw impossible)

        if (opts.verbose >= 5) dbgs("directions = " + directions.mkString(" "))
        
        //val m1 = treeLocs.getCount(t1)
        //val m2 = treeLocs.getCount(t2)

        //// Exit the restaurant
        // Fetch all instances of (t1, t2) and (t3) (this is slow) [update treeLocs]
        val node2set = new java.util.IdentityHashMap[CutNode,CutNode]
        val locs_follow = new ArrayBuffer[CutNode]
        val t12Locs = treeLocs.remove(t1, { node:CutNode =>
          node.followDirections(directions) match {
            case None => false
            case Some(node2) =>
              if (node2.matches(t2)) {
                node2set.put(node2, node2) // Keep track of places where t2 is
                locs_follow += node2
                true
              }
              else
                false
          }
        })
        treeLocs.remove(t2, { node:CutNode => node2set.containsKey(node) })
        val t3Locs = treeLocs.remove(t3, { node:CutNode => true })
        t3Locs.foreach(locs_follow += follow(_))

        // How inefficient are we being?
        //dbgs("EFFICIENCY: %s of %s, %s of %s, %s of %s|%s, %s of %s", m1, t1, m2, t2, t12Locs.size, t1, t2, t3Locs.size, t3)

        val c1 = t1.context
        val c2 = t2.context
        val c3 = t3.context

        params.treeCounts.incr(c1, t1, -t12Locs.size) // [update treeCounts]
        params.treeCounts.incr(c2, t2, -t12Locs.size)
        params.treeCounts.incr(c3, t3, -t3Locs.size)

        //// n total choices to make, need to decide on k
        val locs = (t12Locs ++ t3Locs).toArray
        val n = locs.size

        val alpha1 = params.alphas(c1)
        val alpha2 = params.alphas(c2)
        val alpha3 = params.alphas(c3)
        val p1 = baseProb(t1)
        val p2 = baseProb(t2)
        val p3 = baseProb(t3)
        var nt1 = params.treeCounts.getCount(c1, t1) + alpha1*p1
        var nt2 = params.treeCounts.getCount(c2, t2) + alpha2*p2
        var nt3 = params.treeCounts.getCount(c3, t3) + alpha3*p3
        val nc1 = params.treeCounts.totalCount(c1) + alpha1
        val nc2 = params.treeCounts.totalCount(c2) + alpha2
        val nc3 = params.treeCounts.totalCount(c3) + alpha3
        require (c1 == c3)
        require (t1 != t2)
        //dbgs("%s %s %s", t1, t2, t3)
        //dbgs("%s %s %s", p1, p2, p3)
        if ((nt1 == 0 || nt2 == 0) || nt3 == 0) {
          val defaultValue = 1e-10
          Utils.warnings("All base probabilities are too small, setting to default value of "+defaultValue)
          nt1 = defaultValue
          nt2 = defaultValue
          nt3 = defaultValue
        }

        def computeSlow = {
          val probs = new Array[Double](n+1) // k -> probability of k (t1,t2)s
          foreach(0, n+1, { k:Int =>
            val nd1 = k
            val nd2 = k
            val nd3 = n-k
            probs(k) = crpLogProb(nd1, nt1,                                 nc1) +
                       crpLogProb(nd2, nt2 + I(t2, t1)*nd1,                 nc2 + I(c2, c1)*nd1) + 
                       crpLogProb(nd3, nt3 + I(t3, t1)*nd1 + I(t3, t2)*nd2, nc3 + I(c3, c1)*nd1 + I(c3, c2)*nd2)
          })
          foreach(0, n+1, { k:Int => // Multiply by number of combinations (n choose k)
            probs(k) += logChoose(n, k)
          })

          if (temperature != 1) {
            foreach(0, n+1, { k:Int => // Anneal
              probs(k) /= temperature
            })
          }

          Utils.expNormalizeOrFail_!(probs)
        }

        def computeMed = {
          if (t1 != t2 && c1 != c2) {
            val scores = map(n+1, { k:Int =>
              //logChoose(n, k) + logGenFact(nt1, k) + logGenFact(nt2, k) + logGenFact(nt3, n-k) - logGenFact(nc2, k)
              logGenFact(nt1, k) + logGenFact(nt2, k) + logGenFact(nt3, n-k) - logGenFact(nc2, k) - logGenFact(1, k) - logGenFact(1, n-k)
            })
            //val scores = map(n+1, { k:Int => -logGenFact(1, k) })
            //dbgs("MED SCORES: %s", fmt1(scores))
            Utils.expNormalizeOrFail_!(scores)
          }
          else
            map(n+1, Double.NaN)
        }

        // Supposed to be faster
        def computeFast = {
          val weights = new typesampling.BigDoubleArray(n+1)
          if (t1 == t2) {
            weights.multGenFact(+1, nt1, 2)
          }
          else {
            weights.multGenFact(+1, nt1, 1)
            weights.multGenFact(+1, nt2, 1)
          }
          weights.multGenFactReverse(+1, nt3, 1)
          if (c1 == c2)
            weights.multGenFact(-1, nc1+n, 1) // t1, t2, t3 all draw from c1 (n + k)
          else
            weights.multGenFact(-1, nc2, 1) // t1,t3 draw from c1 (n), t2 draws from c2 (k)
          weights.multGenFactReverse(-1, 1, 1); weights.multGenFact(-1, 1, 1)  // n choose k (ignore n!, which is a constant)
          if (temperature != 1) weights.pow(1.0/temperature)
          //dbgs("FAST SCORES: %s", fmt1(weights.toLog))
          //Utils.expNormalizeOrFail_!(weights.toLog)
          weights.normalize
        }

        //val probs = computeSlow
        val probs = computeFast

        /*val slowProbs = computeSlow
        val medProbs = computeMed
        //val probs = computeFast
        if (Utils.distLinf(slowProbs, probs) > 1e-5) {
          dbgs("================ %s %s", t1 == t2, c1 == c2)
          dbgs("SLOW: %s", fmt1(slowProbs))
          dbgs("MED: %s", fmt1(medProbs))
          dbgs("FAST: %s", fmt1(probs))
        }*/

        //if (opts.verbose >= 5) logs("WEIGHTS: %s", probs.map(math.exp(_)).mkString(" "))

        // Randomly choose k and which k
        val k = Utils.sample(opts.sampleRandom, probs)
        logs("N: %s, K: %s => %s", n, t12Locs.size, k)
        val perm = fig.prob.SampleUtils.samplePermutation(opts.sampleRandom, n)
        val newLocs = fig.basic.ListUtils.applyPermutation(locs, perm)
        val new_t12Locs = newLocs.slice(0, k).toList
        val new_t3Locs = newLocs.slice(k, n).toList
        val newLocs_follow = fig.basic.ListUtils.applyPermutation(locs_follow.toArray:Array[CutNode], perm)
        val new_t12Locs_follow = newLocs_follow.slice(0, k)
        val new_t3Locs_follow = newLocs_follow.slice(k, n)

        //// Re-enter the restaurant
        treeLocs.add(t1, new_t12Locs) // [update treeLocs]
        treeLocs.add(t2, new_t12Locs_follow.toList)
        treeLocs.add(t3, new_t3Locs)

        params.treeCounts.incr(c1, t1, k) // [update treeCounts]
        params.treeCounts.incr(c2, t2, k)
        params.treeCounts.incr(c3, t3, n-k)

        new_t12Locs_follow.foreach { node:CutNode => setCut(node, true) } // [update ex.cut]
        new_t3Locs_follow.foreach { node:CutNode => setCut(node, false) }

        if (opts.verbose >= 5) track("Sampling types: chose k=%s of %s,%s and %s of %s", k, t1, t2, n-k, t3) {
          logs("LOCS: " + locs.mkString(" "))
          logs("NEW LOCS: %s || %s", new_t12Locs.mkString(" "), new_t3Locs.mkString(" "))
          logs("PROBS: %s", fmt1(probs))
          //logs(tmpStateStr + " HAS PROBS: %s", fmt1(probs))
          logs("n=%s locs, p1=%s,p2=%s,p3=%s, nt1=%s,nt2=%s,nt3=%s, nc1=%s,nc2=%s,nc3=%s", n, p1, p2, p3, nt1, nt2, nt3, nc1, nc2, nc3)
          //logState
        }
        /*tmpCounts.incr(oldStateStr + " " + tmpIdx, tmpStateStr, 1)
        logs("TRANS " + oldStateStr + " " + tmpIdx + " TO " + tmpStateStr)
        tmpIdx = (tmpIdx + 1) % 4*/
      }
    }
  }
  def newIterState = new IterState
}
