package typesampling;

import java.io.*;
import java.util.*;
import fig.basic.*;
import static fig.basic.LogInfo.*;
import edu.berkeley.nlp.ling.*;
import edu.berkeley.nlp.util.Filter;

public class TreeUtils {
  // Note: always set Tree.isIntermediateNode = false

  // Can't use the following symbols in the Berkeley Decoder: @, ~, ^, +, =
  private static String intermediateNodePrefix = "_";
  private static String parentAnnotationPrefix = ";";
  private static String regexpParentAnnotationPrefix = ";";
  private static String siblingAnnotationPrefix = "->";
  private static String siblingAnnotationConnective = "-";
  private static String siblingAnnotationEtc = "..";
  private static String unarySeparator = "|";
  private static String regexpUnarySeparator = "\\|";

  private static final HeadFinder headFinder = new CollinsHeadFinder();
  // Binarize direction based on heads
  // First binarize right arguments and then binarize left arguments
  // Note: probabilistically, arguments are generated out to in
  // Example: if the tree is (P L2 L1 H R1 R2) where H is the head, then the result is
  // (@P->..H L2 (@P->..H L1 (@P->H.. (@P->H.. H R1) R2))) if order 
  // order = 0: remember root (e.g., @P and @P)
  // order = 1: remember root and head (e.g., @P->H.. and @P->..H)
  // order = 2: remember root and head and last symbol (e.g., @P->H..Ri and @P->Li..H)
  public static Tree<String> headBinarize(Tree<String> tree, int order) {
    String rootLabel = tree.getLabel();

    if(tree.getChildren().size() == 0)
      return tree;
    if(tree.getChildren().size() == 1)
      return new Tree<String>(rootLabel, false, ListUtils.newList(headBinarize(tree.getChildren().get(0), order)));

    // Recursively binarize all children
    List<Tree<String>> newChildren = new ArrayList();
    for(Tree<String> child : tree.getChildren())
      newChildren.add(headBinarize(child, order));

    int head = tree.getChildren().indexOf(headFinder.determineHead(tree));
    if (head == -1) {
      head = newChildren.size()-1; // Right binarize by default
      throw new RuntimeException("Can't determine head of "+render(tree));
    }
    Tree<String> newTree = newChildren.get(head);
    String headLabel = newChildren.get(head).getLabel();
    int numLeftArgs = head;
    int numRightArgs = newChildren.size()-head-1;
    int numArgs = numLeftArgs + numRightArgs;

    assert (order >= 0 && order <= 2);

    for(int i = 0; i < numRightArgs; i++) { // Right arguments
      boolean isRoot = (--numArgs == 0);
      Tree<String> child = newChildren.get(head+i+1);
      String l = rootLabel;
      if (!isRoot) {
        l = intermediateNodePrefix + l;
             if (order == 1) l += siblingAnnotationPrefix + headLabel + siblingAnnotationEtc;
        else if (order == 2) l += siblingAnnotationPrefix + headLabel + siblingAnnotationEtc + child.getLabel();
      }
      newTree = new Tree<String>(l, false, ListUtils.newList(newTree, child));
    }

    for(int i = 0; i < numLeftArgs; i++) { // Left arguments
      boolean isRoot = (--numArgs == 0);
      Tree<String> child = newChildren.get(head-i-1);
      String l = rootLabel;
      if (!isRoot) {
        l = intermediateNodePrefix + l;
             if (order == 1) l += siblingAnnotationPrefix + siblingAnnotationEtc + headLabel;
        else if (order == 2) l += siblingAnnotationPrefix + child.getLabel() + siblingAnnotationEtc + headLabel;
      }
      newTree = new Tree<String>(l, false, ListUtils.newList(child, newTree));
    }

    return newTree;
  }

  public static Tree<String> parentAnnotate(Tree<String> tree) {
    return parentAnnotate(tree, null);
  }
  public static Tree<String> parentAnnotate(Tree<String> tree, String parentLabel) {
    if (tree.isLeaf()) return tree; // Don't change terminals

    // Recursively label all children
    List<Tree<String>> newChildren = new ArrayList();
    for (Tree<String> child : tree.getChildren())
      newChildren.add(parentAnnotate(child, tree.getLabel()));
    String newLabel = tree.getLabel();
    if (parentLabel != null) newLabel += parentAnnotationPrefix+parentLabel;
    return new Tree<String>(newLabel, false, newChildren);
  }

  // For chain of unaries A -> B -> C, replace with with A|B|C or just C
  public static Tree<String> flattenUnaryChains(Tree<String> tree, boolean annotateChain) {
    if(tree.isLeaf())
      return tree;
    else {
      // Flatten (flatten the chain top down)
      String label = tree.getLabel();
      while (tree.getChildren().size() == 1 && !tree.isPreTerminal()) {
        tree = tree.getChildren().get(0);
        if (annotateChain)
          label += unarySeparator + tree.getLabel();
        else
          label = tree.getLabel();
      }

      // Recurse on children
      List<Tree<String>> children = new ArrayList();
      for(Tree<String> child : tree.getChildren())
        children.add(flattenUnaryChains(child, annotateChain));
      return new Tree<String>(label, false, children);
    }
  }
  public static Tree<String> unflattenUnaryChains(Tree<String> tree) {
    if(tree.isLeaf())
      return tree;
    else {
      String[] labels = tree.getLabel().split(regexpUnarySeparator);
      
      // Recurse on children (start with the bottom of the unary chain)
      List<Tree<String>> children = new ArrayList();
      for(Tree<String> child : tree.getChildren())
        children.add(unflattenUnaryChains(child));
      tree = new Tree<String>(labels[labels.length-1], false, children);
      
      // Unflatten (build the chain bottom up)
      for (int i = labels.length-2; i >= 0; i--) {
        children = new ArrayList();
        children.add(tree);
        tree = new Tree<String>(labels[i], false, children);
      }
      return tree;
    }
  }

  public static Tree<String> removeParentAnnotations(Tree<String> tree) {
    String newLabel = tree.getLabel().replaceAll(regexpParentAnnotationPrefix+".*", "");
    List<Tree<String>> children = new ArrayList();
    for(Tree<String> child : tree.getChildren())
      children.add(removeParentAnnotations(child));
    return new Tree<String>(newLabel, false, children);
  }

  // Take a tree and make it ready for evaluation
  public static Tree<String> polishTree(Tree<String> tree) {
    // Remove intermediate nodes
    //logs("BEFORE: %s", render(tree));
    tree = Trees.spliceNodes(tree, new Filter<String>() {
      public boolean accept(String s) {
        return s.startsWith(intermediateNodePrefix);
      }
    });
    //logs("AFTER REMOVE INTERMEDIATE: %s", render(tree));
    tree = unflattenUnaryChains(tree);
    //logs("AFTER UNFLATTEN UNARY CHAINS: %s", render(tree));
    tree = removeParentAnnotations(tree); // Note that the Berkeley Decoder might already remove these
    //logs("AFTER REMOVING PARENT ANNOTATION: %s", render(tree));
    return tree;
  }

  public static <T> String render(Tree<T> tree) { return Trees.PennTreeRenderer.render(tree); }
}
