package typesampling

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,fail,impossible}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}
import fig.basic.LogInfo.rants

import java.util.{Set,ArrayList,Random}
import fig.exec.Execution.putLogRec
import fig.basic.NumUtils.logGamma
import fig.basic.NumUtils.logAdd
import Misc.logGenFact

// TypeBasedLearner: generic base class for doing type-level sampling.
// Maintains linked list data structure of sites.

////////////////////////////////////////////////////////////

/**
Learns a binomial distribution over m = 0..n conditioned on some covariates x.
Do stochastic gradient.
*/
// Learn how many sites to assign to a particular configuration
class Proposal(opts:LearnOptions) {
  var theta : Array[Double] = null
  var D = -1
  var numUpdates = 0
  val exampleOut = fig.basic.IOUtils.openOutHard(Execution.getFile("proposal.examples"))

  val predRatios = new Accumulator 
  val oracleRatios = new Accumulator 
  val scoreErrors = new Accumulator 

  // Add an example: x are the covariates, probs is the distribution we want to match
  def add(x:Array[Double], probs:Array[Double]) : Unit = {
    if (theta == null) {
      D = x.size
      theta = new Array[Double](D) // Parameters
    }
    val n = probs.size-1
    if (n == 0) return // No decision to make there
    val p = mu(x)
    numUpdates += 1

    val a = Utils.sum(n+1, { m:Int => probs(m) * m })
    val b = n - a
    
    // Try to predict
    if (opts.verbose >= 3) {
      exampleOut.println(fmt(a/(a+b)) + " | " + fmt1(probs) + " | " + fmt1(x))

      def ratio(a:Double, b:Double) = {
        if (a < 1e-10 && b < 1e-10) 1.0 // Don't care about small fractions
        else a/b
      }
      val smoothing = 1e-5
      val score = inv_logistic(p)
      val oracle_p = (a+smoothing)/(a+b+2*smoothing)
      val oracle_score = inv_logistic(oracle_p)
      val oracleProbs = map(n+1, { m:Int => math.exp(logProb(oracle_p, n, m)) })
      val predProbs = map(n+1, { m:Int => math.exp(logProb(p, n, m)) })

      Utils.assertValid(p)
      Utils.assertValid(score)
      Utils.assertValid(oracle_p)
      Utils.assertValid(oracle_score)

      val oracleRatio = Utils.sum(n+1, { m:Int => oracleProbs(m) * (1.0 min ratio(probs(m), oracleProbs(m))) })
      val predRatio = Utils.sum(n+1, { m:Int => predProbs(m) * (1.0 min ratio(probs(m), predProbs(m))) })

      oracleRatios += oracleRatio
      predRatios += predRatio
      scoreErrors += (oracle_score - score).abs

      Utils.track_printAll("Proposal: update #%s", numUpdates) {
        logs("x = %s", fmt1(x))
        logs("theta = %s", fmt1(theta))
        if (opts.verbose >= 4) {
          logs("TRUE: %s", fmt1(probs))
          logs("ORACLE (p = %s, score = %s): %s", oracle_p, oracle_score, fmt1(oracleProbs))
          logs("PRED (p = %s, score = %s): %s", p, score, fmt1(predProbs))
        }
        logs("PERFORMANCE: oracleRatio=%s, predRatio=%s, scoreError=%s", oracleRatios, predRatios, scoreErrors)
        logs("ORACLE RATIO: %s to %s; average acceptance = %s",
          Utils.min(n+1, { m:Int => ratio(probs(m), oracleProbs(m)) }),
          Utils.max(n+1, { m:Int => ratio(probs(m), oracleProbs(m)) }),
          Utils.sum(n+1, { m:Int => oracleProbs(m) * (1.0 min ratio(probs(m), oracleProbs(m))) }))
        logs("PRED RATIO: %s to %s, average acceptance = %s",
          Utils.min(n+1, { m:Int => ratio(probs(m), predProbs(m)) }),
          Utils.max(n+1, { m:Int => ratio(probs(m), predProbs(m)) }),
          Utils.sum(n+1, { m:Int => predProbs(m) * (1.0 min ratio(probs(m), predProbs(m))) }))
      }
    }

    foreach(D, { j:Int =>
      val eta = opts.proposalInitStepSize/math.pow(numUpdates, opts.proposalStepSizeReductionPower)
      theta(j) += eta * x(j) * (a*(1 - p) - b*p) / n
    })
  }

  def sample(random:Random, x:Array[Double], n:Int) : Int = {
    val p = mu(x)
    fig.prob.SampleUtils.sampleBinomial(random, n, p)
  }
  def logProb(x:Array[Double], n:Int, m:Int) : Double = logProb(mu(x), n, m)
  private def logProb(p:Double, n:Int, m:Int) : Double = {
    Misc.logChoose(n, m) + m*math.log(p) + (n-m)*math.log(1-p)
  }

  // Compute the mean parameter of the binomial
  def mu(x:Array[Double]) = logistic(Utils.sum(D, { j:Int => theta(j) * x(j) }))
  def logistic(z:Double) = 1.0/(1 + math.exp(-z))
  def inv_logistic(mu:Double) = {
    if (mu == 0) -10.0
    else if (mu == 1) +10.0
    else math.log(mu/(1-mu))
  }
}

////////////////////////////////////////////////////////////

/**
For type based sampling:
 - An example has many sites.  Choose a site at at random.
 - Identify the site's set of patterns P, used to lookup a list of pattern occurrences.
 - For each pattern occurrence o, we identify the corresponding site s.  Let S be all of these sites.
   Property: that from any s, we would have identified P, and thus the same S.
 - We keep a greedily selected subset so none of the sites "overlap".
   Property: changing one site does not affect the potentials of other sites.
             the likelihood only depends on the number of sites assigned to each of the K possible configurations.

 - For each site s in S:
   * We remove counts from s.
   * Remove all spots around s (and its pattern occurrences).
     Property: changing the configuration of a site of only affects the patterns occurrences on those spots.
 - Sample the configurations of all the sites S.
 - For each site s in S:
   * Add the counts.
   * Add the spots around s.
*/
trait TypeBasedLearner extends Learner {
  type Pattern // [immutable, universal]

  // Where we sample a configuration
  type Site <: AbstractSite
  trait AbstractSite { // [immutable, specific to the current sampling]
    def sameAs_?(that:Site) : Boolean
    def removeNeighboringSpotsPermanently : Unit
    def removeNeighboringSpotsTemporarily : Unit
    def addNeighboringSpots : Unit
  }

  // Linked list of spots where a particular pattern occurs
  trait Occurrence { // [mutable]
    var prev : Occurrence
    var next : Occurrence
  }
  object nullOccurrence extends Occurrence {
    var prev : Occurrence = this
    var next : Occurrence = this
    override def toString = "(null)"
  }
  class HeadOccurrence extends Occurrence {
    var prev : Occurrence = nullOccurrence
    var next : Occurrence = nullOccurrence
    override def toString = "(head) -> "+next
  }
  case class PatternOccurrence(pattern:Pattern, spot:Spot) extends Occurrence {
    var prev : Occurrence = nullOccurrence
    var next : Occurrence = nullOccurrence

    //dbgs("create PatternOccurrence %s %s", pattern, spot)
    if (true) { // Prepend the item to the list (after head)
      val h = patternMap.getOrElseUpdate(pattern, new HeadOccurrence)
      next = h.next; if (h.next != nullOccurrence) h.next.prev = this
      prev = h; h.next = this
    }
    //dbgs("%s", patternMap)

    def destroy = {
      prev.next = next
      if (next != nullOccurrence) next.prev = prev
      if (prev.isInstanceOf[HeadOccurrence] && next == nullOccurrence) // Removed last occurrence of this pattern, so remove it from the map
        patternMap -= pattern
    }

    override def toString = pattern+":"+spot+" -> "+next
  }

  // A spot holds multiple pattern occurrences.
  type Spot <: AbstractSpot // [mutable]
  trait AbstractSpot { self:Spot =>
    // When doing a type-level move, we logically remove spots temporarily, but don't actually free the memory,
    // because there's a good chance that we won't change many of the pattern occurrences
    var removed = false
    val occurrences : Seq[PatternOccurrence]
    def destroy = occurrences.foreach(_.destroy)
  }

  // Map from patterns to list of occurrences
  val patternMap = new HashMap[Pattern, HeadOccurrence]

  type Example <: { // [mutable]
    def log : Unit
    def write(puts:String=>Any) : Unit
    def foreachSite(f:Site => Any) : Unit
    def foreachSpot(f:Spot => Any) : Unit
    def initializeSpots : Unit
    def updateCounts : Unit
    def incrParams(dv:Double) : Unit
  }

  trait SpotSeq {
    def N : Int
    def newSpot(i:Int) : Spot
    def spots : Array[Option[Spot]]
    lazy val lastIterSampled = map(N, { i:Int => -1 })

    // If strict, fail if couldn't perform the operation
    def removeSpotTemporarily(i:Int, strict:Boolean) = {
      spots(i) match {
        case Some(spot) =>
          if (spot.removed && strict) throw fail("Already removed spot at " + i)
          spot.removed = true
        case None => throw impossible
      }
    }
    def removeSpotPermanently(i:Int) = {
      spots(i) match {
        case Some(spot) => spot.destroy; spots(i) = None
        case None =>
      }
    }
    def addSpot(i:Int, strict:Boolean) = {
      spots(i) match {
        case Some(spot) =>
          if (!spot.removed && strict) throw fail("Spot already added at "+i)
          spot.removed = false // Don't need to create a new one (savings!)
        case None =>
          spots(i) = Some(newSpot(i))
      }
    }

    def returnIfSpotExists[A](i:Int, value : =>A) = {
      spots(i) match {
        case None => None
        case Some(spot) => if (spot.removed) None else Some(value)
      }
    }

    def initializeSpots : Unit 
    def foreachSpot(f:Spot=>Any) = {
      foreach(0, N, { i:Int =>
        spots(i) match { case Some(spot) => f(spot); case None => }
      })
    }
  }

  def initExample(ex:Example) : Unit

  override def learnInit = {
    super.learnInit

    Utils.track_foreach("Example", trainExamples, { ex:Example =>
      initExample(ex)
      ex.initializeSpots
      ex.incrParams(1)
      true
    })
  }

  val proposal = new typesampling.Proposal(opts)

  // Tracker
  val choiceSet = new InventoryTracker("Choice", false)
  val moveSet = new InventoryTracker("Move", false)
  val shapeSet = new InventoryTracker("Shape", false)
  //val choiceOut = fig.basic.IOUtils.openOutHard(Execution.getFile("choices"))
  val probsOut = fig.basic.IOUtils.openOutHard(Execution.getFile("probs"))
  val scoresOut = fig.basic.IOUtils.openOutHard(Execution.getFile("scores"))
  //var minSampledProbs = map(100, 10.0)

  abstract class IterState extends super.IterState {
    def newMoveState(site:Site) : Option[MoveState]

    def updateInferenceStats = {
      trainExamples.foreach(_.updateCounts)
    }
    
    def processExampleToken(ex:Example) : Unit = throw fail("Not implemented")

    def processExample(ex:Example) = {
      if (opts.typeLevelSampling)
        ex.foreachSite { site =>
          newMoveState(site) match {
            case None =>
            case Some(state) => state.move
          }
        }
      else
        processExampleToken(ex)
    }

    // Performs a type-level move
    trait MoveState {
      def origSite : Site
      // Return the site for this pattern occurrence return None if it overlaps with existing spots 
      def getSite(o:PatternOccurrence) : Option[Site]

      def numConfigurations : Int
      def getConfiguration(site:Site) : Int
      def setConfiguration(site:Site, c:Int) : Boolean // Return whether something changed

      def incrCounts(c:Int, d:Double) : Unit // Update parameters for a particular configuration
      def incrCounts(counts:Array[Int], d:Double) : Unit = {
        foreach(numConfigurations, { c:Int => incrCounts(c, d*counts(c)) })
      }

      // Find all occurrences of this pattern and remove them (but avoid overlapping patterns)
      def patterns : Seq[Pattern]

      def K = numConfigurations
      def N = sites.size

      val sites = new ArrayBuffer[Site]

      type Scorer <: AbstractScorer
      trait AbstractScorer {
        def indepScore(c:Int) : Double // Equal to computing the score of all N sites having configuration c
        def computeScores(k:Int, n:Int) : Array[Double]

        def localFeatures(k:Int, n:Int) : Array[Double] = throw fail("Not supported") // For training the proposal
        def localScore(k:Int, n:Int, m:Int) : Double // Together forms the score of a count vector
        def localScores(k:Int, n:Int) = map(n+1, { m:Int => localScore(k, n, m) }) // Override if we want to make this faster

        def modelScore(counts:Array[Int]) : Double = modelScore(counts, K, N)
        def modelScore(counts:Array[Int], k:Int, n:Int) : Double = {
          if (k == 0) 0.0
          else {
            val m = counts(k-1)
            localScore(k, n, m) + Misc.logChoose(n, m) + modelScore(counts, k-1, n-m)
          }
        }
      }
      trait AbstractDynProgScorer extends AbstractScorer {
        var sumScores : Array[Array[Double]] = null // k, n -> sum of scores

        // Default implementation: can override if have better way of computing
        def computeScores(k:Int, n:Int) = { // Return m -> score if have m of configuration k-1 
          val scores = localScores(k, n)
          var logChooseSum = 0.0
          foreach(n+1, { m:Int => // choose m of the n to go join m
            if (m > 0) logChooseSum += math.log(1.0*(n-m+1)/m)
            scores(m) += logChooseSum + computeSumScore(k-1, n-m)
          })
          scores
        }

        // Default implementation
        def computeInitScores = map(N+1, { n:Int => localScore(1, n, n) })

        def computeSumScore(k:Int, n:Int) : Double = {
          if (sumScores == null) {
            sumScores = map(K, { c:Int =>
              if (c == 0) computeInitScores
              else map(N+1, { n:Int => Double.NaN })
            })
          }

          var sum = sumScores(k-1)(n)
          if (sum.isNaN) {
            require (k > 1)
            val scores = computeScores(k, n)
            sum = Utils.expNormalizeLogZ_!(scores)
            if (opts.trainProposal) proposal.add(localFeatures(k, n), scores) // Add training point so we can approximate this decision
            sumScores(k-1)(n) = sum
            if (opts.verbose >= 5) dbgs("computeSumScore(k=%s,n=%s) = %s", k, n, sum)
          }
          sum
        }

        // For heuristic stuff
        import HeuristicSearch.{ApproxDistrib,EvalNode}
        var approxDistribs : Array[Array[Option[ApproxDistrib]]] = null // k, n -> approximate distribution over m
        def computeApproxDistrib(k:Int, n:Int) : ApproxDistrib = {
          if (approxDistribs == null)
            approxDistribs = map(K, { c:Int => map(N+1, { n:Int => None }) })

          approxDistribs(k-1)(n) match {
            case Some(distrib) => distrib
            case None =>
              val distrib = {
                if (n == 0) {
                  ApproxDistrib(0, n, EvalNode.invalid)
                }
                else if (k == 1) {
                  val score = localScore(k, n, n)
                  ApproxDistrib(score, n, EvalNode(n, score))
                }
                else {
                  HeuristicSearch.approxDistrib(n+1, opts.proposalPrecision, { m:Int =>
                    localScore(k, n, m) + Misc.logChoose(n, m) + computeApproxDistrib(k-1, n-m).logSum
                  })
                }
              }
              approxDistribs(k-1)(n) = Some(distrib)
              if (opts.verbose >= 5) dbgs("approxDistribs(k=%s,n=%s) = %s", k, n, distrib)
              distrib
          }
        }
      }
      def newScorer : Scorer

      def sampleExact(scorer:Scorer) = {
        require (!opts.hardMoves)
        val newCounts = new Array[Int](K)
        def sample(k:Int, n:Int) : Unit = {
          // NOTE: if (k == 1) produces better results? (see 123.exec on r41)
          if (n == 0) { }
          else if (k == 1)
            newCounts(0) = n
          else {
            val scores = scorer.computeScores(k, n)
            //if (opts.verbose >= 5) dbgs("weights: %s", scores.map(math.exp(_)).mkString(" "))
            val m = Utils.sample(opts.sampleRandom, Utils.expNormalizeOrFail_!(scores))
            if (k == K) {
              if (opts.trainProposal) proposal.add(scorer.localFeatures(k, n), scores) // Add training point so we can approximate this decision in the future
            }

            //choiceSet.add((k, n, m), 0, Misc.fmt1_compress(1e-4, scores))

            newCounts(k-1) = m
            sample(k-1, n-m)
          }
        }
        if (opts.verbose >= 5) track("sample") { sample(K, N) }
        else sample(K, N)
        newCounts
      }

      def sampleConsiderOnlyVertices(oldCounts:Array[Int], scorer:Scorer) = {
        val scores = map(K, { c:Int => scorer.indepScore(c) })
        val c = {
          if (opts.verbose >= 1) dbgs("sampleConsiderOnlyVertices: %s", fmt1(Utils.expNormalizeOrFail_!(scores.toArray)))
          if (opts.hardMoves) Utils.argmax(scores)
          else { Utils.expNormalizeOrFail_!(scores); Utils.sample(opts.sampleRandom, scores) }
        }
        val newCounts = new Array[Int](K)
        if (opts.verbose >= 3) track("hardMove") {
          dbgs("oldCounts = %s", fmt1(oldCounts))
          dbgs("put all n=%s in c=%s", N, c)
          dbgs("fake scores: %s (diff = %s)", fmt1(scores), scores(0)-scores(1))
          dbgs("real scores: %s", fmt1(scorer.computeScores(K, N)))
        }
        newCounts(c) = N
        newCounts
      }

      def sampleGibbsOnSimplex(oldCounts:Array[Int], scorer:Scorer) = {
        // Do a random walk on the simplex of counts
        // FUTURE: make this more efficient
        val counts = oldCounts.toArray
        // TODO
        // Don't see how this will be better than pairs
        //var score = modelScore(counts)
        foreach(100, { q:Int =>
          foreach(K, { c1:Int =>
            foreach(K, { c2:Int =>
              if (c1 != c2) {
                // Try either decrementing or incrementing
                throw impossible
              }
            })
          })
        })
        counts
      }

      // Return a new vector of counts; also return a function to evaluate proposal log probabilities
      def proposeCounts(_scorer:Scorer) : (Array[Int],Array[Int]=>Double) = {
        val scorer = _scorer.asInstanceOf[AbstractDynProgScorer]
        val newCounts = new Array[Int](K)
        def sample(k:Int, n:Int) : Unit = {
          if (n == 0) { }
          else if (k == 1)
            newCounts(0) = n
          else {
            val m = scorer.computeApproxDistrib(k, n).sample(opts.sampleRandom, opts.proposalHeldoutFrac)
            //dbgs("k=%s, n=%s, choose m=%s", k, n, m)
            newCounts(k-1) = m
            sample(k-1, n-m)
          }
        }
        sample(K, N)

        def proposalScore(counts:Array[Int], k:Int, n:Int) : Double = {
          if (n == 0) 0.0
          else if (k == 0) 0.0
          else {
            val m = counts(k-1)
            //dbgs("proposalScore: k=%s, n=%s, m=%s: %s", k, n, m,
              //scorer.approxDistribs(k-1)(n).get.logProb(opts.proposalHeldoutFrac, m) + proposalScore(counts, k-1, n-m))
            scorer.computeApproxDistrib(k, n).logProb(opts.proposalHeldoutFrac, m) + proposalScore(counts, k-1, n-m)
          }
        }
        (newCounts, { counts:Array[Int] => proposalScore(counts, K, N) })
      }

      // Training a proposal didn't pan out
      def proposeCountsUsingTrainedProposal(scorer:Scorer) = {
        val newCounts = new Array[Int](K)
        def sample(k:Int, n:Int) : Unit = {
          if (k == 1)
            newCounts(0) = n
          else {
            val m = proposal.sample(opts.sampleRandom, scorer.localFeatures(k, n), n)
            newCounts(k-1) = m
            sample(k-1, n-m)
          }
        }
        def proposalScore(counts:Array[Int], k:Int, n:Int) : Double = {
          if (k == 0) 0.0
          else {
            val m = counts(k-1)
            proposal.logProb(scorer.localFeatures(k, n), n, m) + proposalScore(counts, k-1, n-m)
          }
        }
        (newCounts, { counts:Array[Int] => proposalScore(counts, K, N) })
      }

      def sampleMH(oldCounts:Array[Int], scorer:Scorer) = track("sampleMH") {
        val (newCounts, proposalScore) = proposeCounts(scorer)
        if (opts.verbose >= 1) dbgs("oldCounts: %s", fmt1(oldCounts))
        if (opts.verbose >= 1) dbgs("newCounts: %s", fmt1(newCounts))

        if (opts.alwaysAccept) newCounts
        else {
          val oldLogP = scorer.modelScore(oldCounts)
          val newLogP = scorer.modelScore(newCounts)
          val oldLogQ = proposalScore(oldCounts)
          val newLogQ = proposalScore(newCounts)
          Utils.assertValid(oldLogP)
          Utils.assertValid(newLogP)
          Utils.assertValid(oldLogQ)
          Utils.assertValid(newLogQ)
          val acceptRatio = math.exp(newLogP + oldLogQ - oldLogP - newLogQ)
          val acceptProb = math.min(1, acceptRatio)
          if (opts.verbose >= 1)
            logs("acceptProb = %s (p: %s -> %s, q: %s -> %s, ratio: %s)", acceptProb, oldLogP, newLogP, oldLogQ, newLogQ, acceptRatio)
          acceptProbFig.add(acceptProb)
          if (opts.sampleRandom.nextDouble < acceptProb)
            newCounts
          else
            oldCounts
        }
      }

      def move = {
        if (opts.verbose >= 5) dbgs("processSite: origSite %s; patterns: %s", origSite, patterns)

        // Get the sites
        val oldCounts = new Array[Int](K)
        def acquireSite(site:Site) = {
          site.removeNeighboringSpotsTemporarily
          oldCounts(getConfiguration(site)) += 1
          sites += site
        }
        acquireSite(origSite) // Make sure we at least sample this site
        require (opts.typeLevelSampling)
        patterns.foreach { pattern => // For each pattern 
          patternMap.get(pattern) match {
            case None =>
            case Some(o) =>
              var ptr : Occurrence = o
              while (ptr.next != nullOccurrence) {
                val o = ptr.next.asInstanceOf[PatternOccurrence]
                getSite(o) match {
                  case Some(candidateSite) =>
                    if (!candidateSite.sameAs_?(origSite)) acquireSite(candidateSite)
                  case None =>
                }
                ptr = ptr.next
              }
          }
        }
        blockSizeFig.add(N)

        incrCounts(oldCounts, -1)

        val scorer = newScorer
        if (opts.verbose >= 3) dbgs("K=%s, N=%s", K, N)

        // Sample new thing!
        val newCounts = {
          if (opts.considerOnlyVertices) sampleConsiderOnlyVertices(oldCounts, scorer)
          else if (opts.gibbsOnSimplex) sampleGibbsOnSimplex(oldCounts, scorer)
          else if (opts.useProposal) sampleMH(oldCounts, scorer)
          else sampleExact(scorer)
        }

        require (Utils.sum(oldCounts) == Utils.sum(newCounts), fmt1(oldCounts) + " => " + fmt1(newCounts))

        //moveSet.add((oldCounts.toList, newCounts.toList))

        // Diagnostics: sampleExact seems to do better exploration than evaluating at the vertices;
        // What are the probabilities under the vertices?
        /*val probs = {
          val scores = map(K, { c:Int => scorer.indepScore(c) })
          Utils.expNormalizeOrFail_!(scores)
          val scores2 = map(K, { c:Int => scorer.modelScore(map(K, { d:Int => if (c == d) N else 0 })) })
          Utils.expNormalizeOrFail_!(scores2)
          if (Utils.distLinf(scores, scores2) > 1e-8) throw fail(fmts("%s || %s", fmt1(scores), fmt1(scores2)))
          scores
        }
        foreach(K, { c:Int =>
          if (newCounts(c) < minSampledProbs.size) {
            if (probs(c) < minSampledProbs(newCounts(c))) {
              minSampledProbs(newCounts(c)) = probs(c) 
              track("Sampled %s of c=%s with indepProb=%s", newCounts(c), c, ""+probs(c)) {
                dbgs("scores: %s", fmt1(map(K, { c:Int => scorer.indepScore(c) })))
                dbgs("scores2: %s", fmt1(map(K, { c:Int => scorer.modelScore(map(K, { d:Int => if (c == d) N else 0 })) })))
                logs("oldCounts: %s", fmt1(oldCounts))
                logs("newCounts: %s", fmt1(newCounts))
              }
            }
          }
          //if (newCounts(c) > 0) require (probs(c) > 0)
        })*/

        // Randomly permute the sites, and then go through in order to assign the sites configurations
        foreach(N, { i:Int =>
          val j = i + opts.sampleRandom.nextInt(N-i)
          Utils.swap(sites, i, j)
        })
        def setAll(k:Int, n:Int) : Unit = {
          if (k > 0) {
            val m = newCounts(k-1)
            foreach(n-m, n, { i:Int =>
              if (setConfiguration(sites(i), k-1))
                sites(i).removeNeighboringSpotsPermanently
            })
            setAll(k-1, n-m)
          }
        }
        setAll(K, N)
        sites.foreach(_.addNeighboringSpots)
        incrCounts(newCounts, +1)
      }
    }
  }

  def repCheck = {
    val occurrences = new HashSet[Occurrence]
    // Go through map from example to occurrence
    trainExamples.foreach { ex:Example =>
      ex.foreachSpot { spot:Spot =>
        require (!spot.removed, spot)
        spot.occurrences.foreach { o =>
          require (o.asInstanceOf[PatternOccurrence].spot == spot)
          occurrences += o
        }
      }
    }
    
    // Go through patternMap
    patternMap.keys.foreach { pattern:Pattern =>
      var o = patternMap(pattern).next
      require (o != nullOccurrence)
      while (o != nullOccurrence) {
        require (occurrences.contains(o))
        val po = o.asInstanceOf[PatternOccurrence]
        require (po.pattern == pattern)
        occurrences -= o
        require (o.prev.next == o)
        if (o.next != nullOccurrence) require (o.next.prev == o)
        o = o.next
      }
    }
    require (occurrences.size == 0)
  }
}
