package typesampling.flatseg

import typesampling._
import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import fig.exec.Execution.putLogRec
import fig.basic.NumUtils.logGamma

////////////////////////////////////////////////////////////

trait FlatSegBase extends typesampling.Learner {
  type Seg = tea.SubArray[String]
  type Example <: ExampleBase

  // Contains elements, predicted segmentation cut, and true segmentation trueCut
  // cut(i) = whether atom i start the beginning of a new segment
  abstract case class ExampleBase(atoms:Array[String], cut:Array[Boolean], trueCut:Array[Boolean]) {
    val cutCounts = new Array[Int](cut.size)
    val uncutCounts = new Array[Int](cut.size)
    val N = atoms.size
    require (cut.size == N+1)
    require (trueCut.size == N+1)
    require (cut(0) == true)
    require (cut(N) == true)

    // Allow efficient traversal of segment boundaries
    val prev = map(N+1, { i:Int => -1 })
    val next = map(N+1, { i:Int => -1 })
    def setCut(i:Int, newCut:Boolean) = { // Return if changed
      //dbgs("setCut i=%s: %s -> %s", i, cut(i), newCut)
      require (i > 0 && i < N)
      if (cut(i)) {
        if (!newCut) {
          // cut -> uncut
          next(prev(i)) = next(i)
          prev(next(i)) = prev(i)
          next(i) = -1
          prev(i) = -1
          cut(i) = newCut
        }
      }
      else {
        if (newCut) {
          prev(i) = compute_prev(i)
          next(i) = compute_next(i)
          next(prev(i)) = i 
          prev(next(i)) = i
          cut(i) = newCut
        }
      }
    }
    // Start of previous and next segments
    def compute_prev(i:Int) = {
      var j = i-1
      while (!cut(j)) j -= 1
      j
    }
    def compute_next(i:Int) = {
      var j = i+1
      while (!cut(j)) j += 1
      j
    }
    def compute_all_prev_next = {
      foreach(0, N+1, { i:Int =>
        if (cut(i)) {
          prev(i) = if (i > 0) compute_prev(i) else -1
          next(i) = if (i < N) compute_next(i) else -1
        }
        else {
          prev(i) = -1
          next(i) = -1
        }
      })
    }

    def numSegs = Utils.sum(N, { i:Int => if (cut(i)) 1 else 0 })

    def incrParams(dv:Double) = {
      foreach(N, { i:Int =>
        if (cut(i)) params.segCounts.incr(rightSeg(i), dv)
      })
    }

    def updateCounts = {
      foreach(N, { i:Int =>
        if (cut(i)) cutCounts(i) += 1
        else uncutCounts(i) += 1
      })
    }

    // Start of previous and next segments
    def get_prev(i:Int) = {
      var j = i-1
      while (!cut(j)) j -= 1
      j
    }
    def get_next(i:Int) = {
      var j = i+1
      while (!cut(j)) j += 1
      j
    }
    def oneSeg(i:Int) = new Seg(atoms, get_prev(i), get_next(i))
    def leftSeg(i:Int) = new Seg(atoms, get_prev(i), i)
    def rightSeg(i:Int) = new Seg(atoms, i, get_next(i))

    // Can be accomplished by leftSeg/rightSeg == seg, but do incremental match for speed
    def leftSegMatches(i:Int, seg:Seg) : Boolean = {
      if (i-seg.size < 0 || !cut(i-seg.size)) false
      else {
        var d = 0
        while (seg.size-d-1 >= 0 && (d == 0 || !cut(i-d)) && atoms(i-d-1) == seg(seg.size-d-1))
          d += 1
        d == seg.size && cut(i-d)
      }
    }
    def rightSegMatches(i:Int, seg:Seg) : Boolean = {
      if (i+seg.size > N || !cut(i+seg.size)) false
      else {
        var d = 0
        while (d < seg.size && (d == 0 || !cut(i+d)) && atoms(i+d) == seg(d))
          d += 1
        d == seg.size && cut(i+d)
      }
    }

    def cutProb(i:Int) : Double = 1.0*cutCounts(i)/(cutCounts(i)+uncutCounts(i))

    def log : Unit = log(true)
    def log(printCounts:Boolean) : Unit = write(logs("%s", _), printCounts)
    def write(puts:String=>Any) : Unit = write(puts, true)
    def write(puts:String=>Any, printCounts:Boolean) : Unit = {
      val predCut = cut
      val trueBuf = new StringBuilder
      val predBuf = new StringBuilder
      foreach(N+1, { i:Int =>
        if (trueCut(i) || predCut(i)) {
          trueBuf.append(if (trueCut(i)) "|" else " ")
          predBuf.append(if (predCut(i)) "|" else " ")
        }
        if (i < N) {
          trueBuf.append(atoms(i))
          predBuf.append(atoms(i))
        }
      })
      puts(fmts("True: %s", trueBuf))
      puts(fmts("Pred: %s", predBuf))

      if (printCounts) {
        val s = map(N, { i:Int =>
          //(if (i == 0) "" else "["+cutCounts(i)+","+uncutCounts(i)+"]")+atoms(i)
          (if (i == 0) "" else fmts("[%s]", cutProb(i)))+atoms(i)
        }).mkString("")
        puts(fmts("Counts: %s", s))
      }
    }
  }
  def newExample(atoms:Array[String], cut:Array[Boolean], trueCut:Array[Boolean]) : Example

  def initExample(ex:Example) = {
    opts.initType match {
      case Enums.InitType.random =>
        foreach(1, ex.N, { i:Int => ex.cut(i) = opts.initRandom.nextDouble < opts.initCutProb })
    }
    ex.compute_all_prev_next
  }

  // The node is left branching and forms a list of segments.
  // Each segment node is left branching and forms a list of atoms.
  def convertNode(node:Node) = {
    val segs = new ArrayBuffer[String]
    val trueCut = new ArrayBuffer[Boolean]
    def convertAtoms(node:Node) : Unit = {
      val i = segs.size
      node match {
        case node:PrimitiveNode =>
          segs += node.value.asInstanceOf[String]
          trueCut += false
        case node:CompositeNode =>
          convertAtoms(node.left)
          convertAtoms(node.right)
      }
      trueCut(i) |= node.context.asInstanceOf[PrimitiveContext].value.asInstanceOf[Boolean]
    }
    convertAtoms(node)
    trueCut += true
    trueCut(0) = true

    val cut = new Array[Boolean](trueCut.size)
    cut(0) = true; cut(cut.size-1) = true
    newExample(segs.toArray, cut, trueCut.toArray)
  }

  class Params extends AbstractParams {
    var alpha = opts.alpha
    var atomContProb = opts.initBaseContProb
    var segContProb = opts.initBaseContProb
    val atomCounts = new CounterMap[String]
    val segCounts = new CounterMap[Seg]

    def write(maxPrint:Int, puts:String=>Any) = {
      puts("ALPHA\t"+fmt(alpha))
      puts("CHAR_CONT\t"+fmt(atomContProb))
      puts("WORD_CONT\t"+fmt(segContProb))
      atomCounts.write(maxPrint, true, "CHAR\t", puts)
      segCounts.write(maxPrint, true, "WORD\t", puts)
    }

    def size = segCounts.numNonZero
    def incr(that:Params) = {
      this.segCounts.incr(1, that.segCounts)
    }
  }
  val params = new Params
  val avgParams = new Params

  trait IterStateBase extends IterState {
    override def log = {
      super.log
      val numSegs = Utils.sum(trainExamples.size, { i:Int => trainExamples(i).numSegs })
      putLogRec("totalSegs", numSegs)
      putLogRec("avgSegs", 1.0*numSegs/trainExamples.size)
    }

    override def sampleHyperparameters = {
      if (opts.sampleAlpha)
        params.alpha = sampleAlpha(params.segCounts)

      if (opts.sampleBaseContProb) { // segContProb
        var numCont = 0
        var numStop = 0
        trainExamples.foreach { ex:Example =>
          numCont += ex.numSegs - 1
          numStop += 1
        }
        params.segContProb = sampleContProb(numCont, numStop)
      }
      if (opts.sampleBaseContProb) { // atomContProb
        var numCont = 0
        var numStop = 0
        params.segCounts.keys.foreach { s:Seg =>
          numCont += s.size - 1
          numStop += 1
        }
        params.atomContProb = sampleContProb(numCont, numStop)
      }
    }

    def setCut(ex:Example, i:Int, newCut:Boolean) : Boolean = {
      numSampledVars += 1
      if (ex.cut(i) != newCut) {
        ex.cut(i) = newCut
        numChangedVars += 1
        true
      }
      else
        false
    }
  }

  def stateLogProb(segCounts:CounterMap[Seg]) = {
    var logProb = 0.0
    logProb += margLogProb(segCounts, params.alpha, baseLogProb _)

    // Probability of generating a certain number of segments
    trainExamples.foreach { ex:Example =>
      val numSegs = ex.numSegs
      if (numSegs > 1) logProb += (numSegs-1) * math.log(params.segContProb) // Continue
      logProb += math.log(1-params.segContProb) // Stop
    }
    Utils.assertValid(logProb)
    logProb
  }
  def stateLogProb : Double = stateLogProb(params.segCounts)

  override def stateAccuracy : Double = {
    stateCutAccuracy // Just use this to print stuff out
    stateF1
  }

  def stateF1 : Double = {
    def getSegs(cut:Array[Boolean], posteriors:Array[Double]) = {
      val segs = new java.util.HashSet[(Int,Int)]
      var start = 0
      foreach(1, cut.size, { i:Int =>
        // Posterior decoding: awful idea
        if (if (opts.usePosteriorDecoding) posteriors(i) > opts.posteriorThreshold else cut(i)) { // Beginning of segment
          segs.add((start,i))
          start = i
        }
      })
      segs
    }
    val result = new fig.basic.EvalResult
    trainExamples.foreach { ex:Example =>
      val trueSegs = getSegs(ex.trueCut, map(ex.N+1, { i:Int => if (ex.trueCut(i)) 1.0 else 0.0 }))
      val predSegs = getSegs(ex.cut, map(ex.N+1, { i:Int => ex.cutProb(i) }))
      result.add(trueSegs, predSegs)
    }
    logs("Performance: %s", result)
    result.f1
  }
  def stateCutAccuracy : Double = {
    // For this precision/recall tells you about oversegmenting/undersegmentation
    val result = new fig.basic.EvalResult
    trainExamples.foreach { ex:Example =>
      foreach(1, ex.N, { i:Int =>
        result.add(ex.trueCut(i), ex.cut(i))
      })
    }
    logs("Cut performance: %s", result)
    result.f1
  }

  // Probability of generating this tree from the base distribution
  // At each node (except the root), we choose whether to stop or not.
  def baseProb(seg:Seg) : Double = {
    var p = 1.0
    foreach(seg.size, { i:Int =>
      p *= params.atomCounts.getProb(seg(i)) *
        (if (i == seg.size-1) 1-params.atomContProb else params.atomContProb)
    })
    p
  }
  def baseLogProb(seg:Seg) : Double = {
    var p = 0.0
    foreach(seg.size, { i:Int =>
      p += math.log(params.atomCounts.getProb(seg(i)) *
        (if (i == seg.size-1) 1-params.atomContProb else params.atomContProb))
    })
    p
  }

  def repCheck = {
    def change(dv:Double) = trainExamples.foreach(_.incrParams(dv))
    change(-1)
    require (params.segCounts.numNonZero == 0)
    change(+1)
  }

  override def stateStr(ex:Example) = ex.cut.mkString(" ")
  def getInferenceStats(ex:Example) = {
    map(ex.N, { i:Int => ex.cutProb(i) })
  }
}

////////////////////////////////////////////////////////////

// Given a sequence of atoms (strings), group them into segments
case class FlatSegLearner(opts:LearnOptions) extends typesampling.Learner with FlatSegBase {
  // For EM
  var emLogZ = 0.0
  val newSegCounts = new CounterMap[Seg]

  class Example(atoms:Array[String], cut:Array[Boolean], trueCut:Array[Boolean]) extends ExampleBase(atoms, cut, trueCut)
  def newExample(atoms:Array[String], cut:Array[Boolean], trueCut:Array[Boolean]) = new Example(atoms, cut, trueCut)

  override def learnInit = {
    super.learnInit

    Utils.track_foreach("Example", trainExamples, { ex:Example =>
      initExample(ex)
      foreach(ex.N, { i:Int => params.atomCounts.incr(ex.atoms(i), 1) })

      if (opts.em && opts.uniformInitParams) {
        // Uniform over word types: does much better (on CHILDES, 79% with, 75% without)
        def upper(i:Int) = math.min(ex.N, i+opts.maxSegmentLength)
        foreach(ex.N, { i:Int =>
          foreach(i+1, upper(i)+1, { j:Int =>
            val s = new Seg(ex.atoms, i, j)
            if (params.segCounts.getCount(s) == 0)
              params.segCounts.incr(s, 1)
          })
        })
      }
      else {
        foreach(ex.N, { i:Int => if (ex.cut(i)) params.segCounts.incr(ex.rightSeg(i), 1) })
      }

      if (opts.typeLevelSampling) {
        foreach(ex.N, { i:Int =>
          if (ex.cut(i))
            segLocs.add(ex.rightSeg(i), Loc(ex, i) :: Nil) // Remember where this segment is
        })
      }

      true
    })
  }

  ////////////////////////////////////////////////////////////

  class IterState extends super.IterState with IterStateBase {
    def processExample(ex:Example) = {
      require (!opts.sampleSiteOncePerIter)
      if (opts.sampleRandom.nextDouble < opts.dynamicProgramFrac)
        sampleCutsTogether(ex)
      else
        sampleCutsSeparately(ex)
    }

    def sampleCutsSeparately(ex:Example) = {
      foreach(1, ex.N, { i:Int =>
        if (opts.verbose >= 2) Utils.track_printAll("State") { ex.log(true) }
        sampleCut(ex, i, burnedIn_?(iter))
      })
    }

    // Sample all cuts in an example using a dynamic program (sentence-based block sampling)
    def sampleCutsTogether(ex:Example) = {
      // Exit restaurant
      if (!opts.em) {
        foreach(ex.N, { i:Int =>
          if (ex.cut(i)) {
            val s = ex.rightSeg(i)
            params.segCounts.incr(s, -1)
            if (opts.typeLevelSampling)
              segLocs.remove(s, Loc(ex, i))
          }
        })
      }

      def seg(i:Int, j:Int) = new Seg(ex.atoms, i, j)

      // TODO: continuation probabilities
      def potential(i:Int, j:Int) = {
        val s = seg(i, j)
        val weight = {
          if (opts.em) params.segCounts.getProb(s)+1e-30
          else (params.segCounts.getCount(s) + params.alpha * baseProb(s)) / (params.segCounts.totalCount + params.alpha) * params.segContProb
        }
        if (opts.useDoubleExpPrior)
          weight * math.exp(-math.pow(s.size, opts.doubleExpPriorRate))
        else
          weight
      }
      val n = ex.N
      def upper(i:Int) = math.min(n, i+opts.maxSegmentLength)
      def lower(i:Int) = math.max(0, i-opts.maxSegmentLength)
      def computeScore(logInside:Array[Double], i:Int, j:Int) = {
        math.log(potential(i, j)) + logInside(j)
      }
      def computeScores(logInside:Array[Double], i:Int) = {
        map(i+1, upper(i)+1, { j:Int => computeScore(logInside, i, j) }) // First segment is [i,j)
      }

      // Find the posterior distribution over all segmentations of the sequence,
      // where each segment has limited length
      val logInside = new Array[Double](n+1) // i -> sum starting with i
      logInside(n) = 0
      Utils.foreachReverse(n, { i:Int =>
        logInside(i) = Utils.expNormalizeLogZ_!(computeScores(logInside, i))
      })

      if (opts.em) { // EM
        val logOutside = new Array[Double](n+1) // i -> sum ending with i-1
        val logZ = logInside(0)
        emLogZ += logZ
        logOutside(0) = 0
        foreach(1, n+1, { i:Int => // Ending i
          val scores = map(lower(i), i, { j:Int => logOutside(j) + math.log(potential(j, i)) }) // Last segment is [j,i)
          logOutside(i) = Utils.expNormalizeLogZ_!(scores)
        })
        // Update new counts
        foreach(n, { i:Int =>
          foreach(i+1, upper(i)+1, { j:Int =>
            val post = math.exp(logOutside(i) + math.log(potential(i, j)) + logInside(j) - logZ)
            Utils.assertValid(post)
            newSegCounts.incr(seg(i, j), post)
          })
        })

        // Find the best segmentation
        val bestSores = new Array[Double](n+1) // i -> sum starting with i
        bestSores(n) = 0
        Utils.foreachReverse(n, { i:Int => bestSores(i) = Utils.max(computeScores(bestSores, i)) })

        var i = 0
        while (i < n) {
          val scores = computeScores(bestSores, i)
          val j = i+1 + Utils.argmax(scores)
          //dbgs("choose %s", j)
          ex.cut(i) = true
          foreach(i+1, j, { k:Int => ex.cut(k) = false })

          if (opts.hardMoves)
            newSegCounts.incr(new Seg(ex.atoms, i, j), 1)

          i = j
        }
      }
      else { // Sample a segmentation
        require (!opts.hardMoves)

        def proposalLogProb(cut:Array[Boolean]) = {
          def recurse(i:Int, j:Int) : Double = {
            if (i == n) 0.0
            else if (cut(j))
              (computeScore(logInside, i, j) - logInside(i)) + recurse(j, j+1)
            else
              recurse(i, j+1)
          }
          recurse(0, 1)
        }

        def modelLogProb(cut:Array[Boolean]) = {
          val newIncrCounts = new CounterMap[Seg]
          def recurse(i:Int, j:Int) : Double = {
            if (i == n) 0.0
            else if (cut(j)) {
              val s = seg(i, j)
              val score = math.log((params.segCounts.getCount(s) + newIncrCounts.getCount(s) + params.alpha * baseProb(s)) /
              (params.segCounts.totalCount + newIncrCounts.totalCount + params.alpha) * params.segContProb)
              newIncrCounts.incr(s, 1)
              score + recurse(j, j+1)
            }
            else
              recurse(i, j+1)
          }
          recurse(0, 1)
        }

        // Sample newCut
        val newCut = new Array[Boolean](n+1)
        var i = 0
        while (i < n) {
          newCut(i) = true
          val probs = computeScores(logInside, i)
          Utils.expNormalizeOrFail_!(probs)
          val j = i+1 + Utils.sample(opts.sampleRandom, probs)
          i = j
        }
        newCut(n) = true

        // Decide accept/reject
        var oldLogQ = proposalLogProb(ex.cut)
        var oldLogP = modelLogProb(ex.cut)
        var newLogQ = proposalLogProb(newCut)
        var newLogP = modelLogProb(newCut)
        val acceptProb = math.min(1, math.exp(newLogP + oldLogQ - oldLogP - newLogQ))
        if (opts.verbose >= 3)
          logs("acceptProb = %s (p: %s -> %s, q: %s -> %s)", acceptProb, oldLogP, newLogP, oldLogQ, oldLogQ)
        if (!opts.alwaysAccept && opts.sampleRandom.nextDouble >= acceptProb) // Reject
          foreach(n, { i:Int => newCut(i) = ex.cut(i) })
        acceptProbFig.add(acceptProb)

        // Move to new configuration
        foreach(n, { i:Int => setCut(ex, i, newCut(i)) })
        foreach(n, { i:Int =>
          if (ex.cut(i)) {
            val s = ex.rightSeg(i)
            params.segCounts.incr(s, 1)
            if (opts.typeLevelSampling)
              segLocs.add(s, Loc(ex, i))
          }
        })
      }
    }

    def updateInferenceStats = {
      trainExamples.foreach(_.updateCounts)
      avgParams.incr(params)
    }

    override def run = {
      if (opts.em) {
        newSegCounts.clear
        emLogZ = 0.0
      }
      super.run
    }

    override def sampleHyperparameters = {
      super.sampleHyperparameters
      if (opts.em) { // M-step
        params.segCounts.clear
        params.segCounts.incr(1, newSegCounts)
      }
    }

    def sampleCut(ex:Example, i:Int, burnedIn:Boolean) = {
      if (opts.typeLevelSampling) sampleCutTypeLevel(ex, i, burnedIn)
      else sampleCutTokenLevel(ex, i, burnedIn)
    }

    def sampleCutTypeLevel(ex:Example, i:Int, burnedIn:Boolean) = {
      if (opts.multiType) sampleCutMultiTypeLevel(ex, i, burnedIn)
      else sampleCutUniTypeLevel(ex, i, burnedIn)
    }

    def sampleCutUniTypeLevel(ex:Example, i:Int, burnedIn:Boolean) = {
      val s1 = ex.leftSeg(i)
      val s2 = ex.rightSeg(i)
      val s3 = ex.oneSeg(i)

      if (s1 == s2) {
        logs("Got %s, deferring to token level sampling", s3)
        sampleCutTokenLevel(ex, i, burnedIn)
      }
      else {
        // Three things to update: segLocs, segCounts, ex.cut
        //val m1 = segLocs.getCount(s1)
        //val m2 = segLocs.getCount(s2)

        //// Exit the restaurant
        // Fetch all instances of (s1, s2) and (s3) (this is slow) [update segLocs]

        //val s12Locs = segLocs.remove(s1, { loc:Loc => loc.i+s1.size < loc.ex.N && loc.ex.rightSeg(loc.i+s1.size) == s2 }) // SLOW
        val s12Locs = segLocs.remove(s1, { loc:Loc => loc.ex.rightSegMatches(loc.i+s1.size, s2) }) // FAST
        //val s12Locs = segLocs.remove(s1, { loc:Loc => Utils.same(loc.ex.rightSegMatches(loc.i+s1.size, s2), loc.i+s1.size < loc.ex.N && loc.ex.rightSeg(loc.i+s1.size) == s2) }) // CHECK

        //segLocs.remove(s2, { loc:Loc => loc.i > 0 && loc.ex.leftSeg(loc.i) == s1}) // SLOW
        segLocs.remove(s2, { loc:Loc => loc.ex.leftSegMatches(loc.i, s1) }) // FAST
        //segLocs.remove(s2, { loc:Loc => Utils.same(loc.ex.leftSegMatches(loc.i, s1), loc.i > 0 && loc.ex.leftSeg(loc.i) == s1) }) // CHECK

        val s3Locs = segLocs.remove(s3, { loc:Loc => true })

        // How inefficient are we being?
        //dbgs("EFFICIENCY: %s of %s, %s of %s, %s of %s|%s, %s of %s", m1, s1, m2, s2, s12Locs.size, s1, s2, s3Locs.size, s3)

        params.segCounts.incr(s1, -s12Locs.size) // [update segCounts]
        params.segCounts.incr(s2, -s12Locs.size)
        params.segCounts.incr(s3, -s3Locs.size)

        //// n total choices to make, need to decide on k
        val locs = (s12Locs ++ s3Locs).toArray
        val n = locs.size
        val probs = new Array[Double](n+1) // k -> probability of k (s1,s2)s

        val de12 = -math.pow(s1.size, opts.doubleExpPriorRate) - math.pow(s1.size, opts.doubleExpPriorRate)
        val de3 = -math.pow(s3.size, opts.doubleExpPriorRate)

        val b1 = params.alpha*baseProb(s1) + params.segCounts.getCount(s1)
        val b2 = params.alpha*baseProb(s2) + params.segCounts.getCount(s2)
        val b3 = params.alpha*baseProb(s3) + params.segCounts.getCount(s3)
        val B  = params.alpha              + params.segCounts.totalCount
        val logContProb = math.log(params.segContProb)
        var sum = 0.0
        foreach(1, n+1, { m:Int => // m is the number of cuts
          // Order of operations is important (m-1+b1 != b1+m-1 when m = 1 and b1 = 1e-30)
          sum += math.log(((b1+(m-1)) * (b2+(m-1)) / (B+(n+m-1)) / (b3+(n-m))) * (n-m+1) / m) + logContProb
          if (opts.useDoubleExpPrior) sum += de12 - de3 // This sucks
          Utils.assertValid(sum)
          probs(m) = sum
        })

        if (temperature != 1) {
          foreach(0, n+1, { k:Int => // Anneal
            probs(k) /= temperature
          })
        }
        Utils.expNormalizeOrFail_!(probs)

        // Randomly choose k and which k
        val k = Utils.sample(opts.sampleRandom, probs)
        val newLocs = fig.basic.ListUtils.applyPermutation(locs, fig.prob.SampleUtils.samplePermutation(opts.sampleRandom, n))
        val new_s12Locs = newLocs.slice(0, k).toList
        val new_s3Locs = newLocs.slice(k, n).toList

        //// Re-enter the restaurant
        segLocs.add(s1, new_s12Locs) // [update segLocs]
        segLocs.add(s2, new_s12Locs.map(_.incr(s1.size)))
        segLocs.add(s3, new_s3Locs)

        params.segCounts.incr(s1, k) // [update segCounts]
        params.segCounts.incr(s2, k)
        params.segCounts.incr(s3, n-k)

        //if (special) logs("CUT %s", new_s12Locs.map(_.i).mkString(" "))

        new_s12Locs.foreach { loc:Loc => setCut(loc.ex, loc.i+s1.size, true) } // [update ex.cut]
        new_s3Locs.foreach { loc:Loc => setCut(loc.ex, loc.i+s1.size, false) }

        if (opts.verbose >= 2) track("Sampling types: chose %s of %s,%s and %s of %s", k, s1, s2, n-k, s3) {
          logs("LOCS: " + locs.mkString(" "))
          logs("NEW LOCS: %s || %s", new_s12Locs.mkString(" "), new_s3Locs.mkString(" "))
          logs("PROBS: %s", probs.mkString(" "))
          //logs("n=%s locs, p1=%s,p2=%s,p3=%s, n1=%s,n2=%s,n3=%s", n, p1, p2, p3, n1, n2, n3)
        }
      }
    }

    def sampleCutMultiTypeLevel(ex:Example, i:Int, burnedIn:Boolean) = {
      val s1 = ex.leftSeg(i)
      //val s2 = ex.rightSeg(i)
      //val s3 = ex.oneSeg(i)
      // UNFINISHED

      if (s1 == ex.rightSeg(i)) {
        logs("Got %s, deferring to token level sampling", ex.oneSeg(i))
        sampleCutTokenLevel(ex, i, burnedIn)
      }
      else {
        // Three things to update: segLocs, segCounts, ex.cut

        //// Exit the restaurant
        // Fetch all instances of (s1, s2) for all s2 [update segLocs]
        // For each of these locations, see what's to the right of s1
        val s2Map = new HashMap[Seg, List[Loc]] // s2 -> list of locations where (s1, s2) occurs
        segLocs.remove(s1, { loc:Loc => loc.i+s1.size < loc.ex.N && loc.ex.rightSeg(loc.i+s1.size) != s1 }).foreach { loc:Loc =>
          val s2 = loc.ex.rightSeg(loc.i+s1.size)
          Misc.addToList(s2Map, s2, loc)
        }
        val s2s = s2Map.keys.toList.toArray
        val M = s2s.size
        val s12Locs = s2s.map(s2Map(_))
        val s3s = s12Locs.map { locs:List[Loc] =>
          val loc = locs.head
          loc.ex.oneSeg(loc.i+s1.size)
        }
        val s3Locs = s3s.map { s3 => segLocs.remove(s3, { loc:Loc => true }) }

        if (opts.verbose >= 2) track("Sampling multi-types") {
          logs("s1 = %s", s1)
          logs("sb = %s", map(M, { m:Int => s2s(m)+":"+s12Locs(m).size+","+s3Locs(m).size }))
          // TODO
        }
      }
    }

    def sampleCutTokenLevel(ex:Example, i:Int, burnedIn:Boolean) = {
      // Consider cutting or uncutting this node
      // If cut, then get two segments (s1, s2)
      // If uncut, then get one segment (s3)
      val s1 = ex.leftSeg(i)
      val s2 = ex.rightSeg(i)
      val s3 = ex.oneSeg(i)

      // Token-level sampling
      if (ex.cut(i)) {
        params.segCounts.incr(s1, -1)
        params.segCounts.incr(s2, -1)
        if (opts.typeLevelSampling) {
          segLocs.remove(s1, Loc(ex, i-s1.size))
          segLocs.remove(s2, Loc(ex, i))
        }
      }
      else {
        params.segCounts.incr(s3, -1)
        if (opts.typeLevelSampling)
          segLocs.remove(s3, Loc(ex, i-s1.size))
      }

      def I[T](x:T,y:T) = (if (x == y) 1 else 0)

      // p(s1, s2 | ...)
      val cut1Weight = (params.segCounts.getCount(s1) + params.alpha * baseProb(s1)) /
                       (params.segCounts.totalCount + params.alpha)
      val cut2Weight = (params.segCounts.getCount(s2) + I(s1, s2) + params.alpha * baseProb(s2)) /
                       (params.segCounts.totalCount + 1 + params.alpha)
      val cutWeight = cut1Weight * params.segContProb * cut2Weight
      // p(s3 | ...)
      val uncutWeight = (params.segCounts.getCount(s3) + params.alpha * baseProb(s3)) /
                        (params.segCounts.totalCount + params.alpha)
      
      val cutProb = annealProb(cutWeight, uncutWeight)
      val uncutProb = 1 - cutProb

      setCut(ex, i, opts.sampleRandom.nextDouble < cutProb)
      if (ex.cut(i)) {
        params.segCounts.incr(s1, 1)
        params.segCounts.incr(s2, 1)
        if (opts.typeLevelSampling) {
          segLocs.add(s1, Loc(ex, i-s1.size))
          segLocs.add(s2, Loc(ex, i))
        }
      }
      else {
        params.segCounts.incr(s3, 1)
        if (opts.typeLevelSampling)
          segLocs.add(s3, Loc(ex, i-s1.size))
      }

      if (opts.verbose >= 2) track("SAMPLE at position i=%s", i) {
        logs("CUT %s * %s = %s (prob %s): %s AND %s", cut1Weight, cut2Weight, cutWeight, cutProb, s1, s2)
        logs("UNCUT %s (prob %s): %s", uncutWeight, uncutProb, s3)
        logs("CHOOSE cut = %s", ex.cut(i))
        //if (opts.verbose >= 3) writeParams(logs("%s", _))
      }
    }
  }
  def newIterState = new IterState

  override def stateLogProb : Double = {
    if (opts.em) {
      // Rebuild segCounts 
      val counts = new CounterMap[Seg]
      trainExamples.foreach { ex =>
        foreach(ex.N, { i:Int =>
          if (ex.cut(i)) counts.incr(ex.rightSeg(i), 1)
        })
      }
      stateLogProb(counts)
    }
    else
      super.stateLogProb
  }

  override def stateAccuracy = {
    if (opts.em)
      logs("emLogZ = %s", emLogZ/Utils.sum(trainExamples.size, { i:Int => trainExamples(i).N }))
    super.stateAccuracy
  }

  override def repCheck : Unit = {
    super.repCheck

    // Check that segLocs is consistent with segCounts
    if (opts.typeLevelSampling) {
      require (segLocs.map.size == params.segCounts.numNonZero, segLocs + " " + params.segCounts)
      segLocs.map.keys.foreach { seg:Seg =>
        val locs = segLocs.map(seg)
        require (locs.size == params.segCounts.getCount(seg))
        segLocs.map(seg).foreach { loc:Loc =>
          require (loc.ex.cut(loc.i))
          require (loc.ex.rightSeg(loc.i) == seg)
        }
      }
    }
  }

  // For type-level sampling
  case class Loc(ex:Example, i:Int) {
    def incr(di:Int) = Loc(ex, i+di)
    override def toString = "["+i+"]"
  }
  val segLocs = new ListHashMap[Seg,Loc]
}
