package typesampling.hmm

import typesampling._
import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,fail,impossible}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import fig.exec.Execution.putLogRec
import fig.basic.NumUtils.logGamma
import Misc.logGenFact
import fig.basic.Indexer

case class HMMLearner(opts:LearnOptions) extends typesampling.TypeBasedLearner {
  val K = opts.numStates
  require (K > 0)

  case class Pattern(prevState:Int, currWord:String, nextState:Int)

  type Site = Spot
  case class Spot(ex:Example, i:Int) extends AbstractSpot with AbstractSite {
    // Site members
    def sameAs_?(that:Site) = (this.ex eq that.ex) && this.i == that.i
    def removeNeighboringSpotsPermanently = {
      ex.removeSpotPermanently(i)
      if (i-1 >= 0) ex.removeSpotPermanently(i-1)
      if (i+1 < ex.N) ex.removeSpotPermanently(i+1)
    }
    def removeNeighboringSpotsTemporarily = {
      ex.removeSpotTemporarily(i, true)
      if (i-1 >= 0) ex.removeSpotTemporarily(i-1, false)
      if (i+1 < ex.N) ex.removeSpotTemporarily(i+1, false)
    }
    def addNeighboringSpots = {
      ex.addSpot(i, true)
      if (i-1 >= 0) ex.addSpot(i-1, false)
      if (i+1 < ex.N) ex.addSpot(i+1, false)
    }

    // Spot members
    val occurrences = PatternOccurrence(Pattern(ex.getState(i-1), ex.words(i), ex.getState(i+1)), this) :: Nil
    override def toString = "["+i+"]"
  }

  case class Example(words:Array[String], states:Array[Int], trueTags:Array[String]) extends SpotSeq {
    val N = words.size
    val counts = map(N, { i:Int => new Array[Int](K) })
    var totalCount = 0

    val spots : Array[Option[Spot]] = map(N, { i:Int => None })
    def newSpot(i:Int) = Spot(this, i)
    def initializeSpots = foreach(N, { i:Int => addSpot(i, true) })

    def foreachSite(f:Site=>Any) = foreach(N, { i:Int =>
      if (!opts.sampleSiteOncePerIter || iter != lastIterSampled(i))
        f(spots(i).get)
    })

    def getState(i:Int) = if (i == -1 || i == N) -1 else states(i)
    def updateCounts = {
      foreach(N, { i:Int => counts(i)(states(i)) += 1 })
      totalCount += 1
    }

    def incrParams(dv:Double) = {
      foreach(N, { i:Int =>
        if (i == 0)
          params.startCounts.incr(states(i), dv)
        else
          params.transitionCounts(states(i-1)).incr(states(i), dv)
        params.emissionCounts(states(i)).incr(words(i), dv)
      })
    }

    def log : Unit = write(logs("%s", _))
    def write(puts:String=>Any) : Unit = {
      // TODO
    }
  }

  // The node is left branching and forms a list of segments.
  // Each segment node is left branching and forms a list of words.
  def convertNode(rootNode:Node) = {
    // Assume we've read in a parse tree where the terminals are words and preterminals are the tags
    val words = new ArrayBuffer[String]
    val trueTags = new ArrayBuffer[String]
    rootNode.foreachNode { node:Node =>
      node match {
        case node:PrimitiveNode =>
          words += node.value.asInstanceOf[String]
          trueTags += node.context.asInstanceOf[PrimitiveContext].value.toString
        case node:CompositeNode =>
      }
    }
    Example(words.toArray, new Array[Int](words.size), trueTags.toArray)
  }

  class Params extends AbstractParams {
    var startAlpha = opts.alpha
    var transitionAlpha = opts.alpha
    var emissionAlpha = opts.alpha
    val startCounts = new CounterVec(K)
    val transitionCounts = map(K, { a:Int => new CounterVec(K) })
    val emissionCounts = map(K, { a:Int => new CounterMap[String] })

    def write(maxPrint:Int, puts:String=>Any) = {
      puts("START_ALPHA\t"+fmt(startAlpha))
      puts("TRANS_ALPHA\t"+fmt(transitionAlpha))
      puts("EMIT_ALPHA\t"+fmt(emissionAlpha))
      startCounts.write(maxPrint, true, "START\t", puts)
      foreach(K, { a:Int =>
        transitionCounts(a).write(maxPrint, true, "TRANS\t"+a+"\t", puts)
      })
      foreach(K, { a:Int =>
        emissionCounts(a).write(maxPrint, true, "EMIT\t"+a+"\t", puts)
      })
    }

    def size = {
      startCounts.numNonZero +
      Utils.sum(K, { a:Int => transitionCounts(a).numNonZero }) +
      Utils.sum(K, { a:Int => emissionCounts(a).numNonZero })
    }

    def incr(that:Params) = {
      this.startCounts.incr(1, that.startCounts)
      foreach(K, { a:Int =>
        this.transitionCounts(a).incr(1, that.transitionCounts(a))
        this.emissionCounts(a).incr(1, that.emissionCounts(a))
      })
    }
  }
  val params = new Params
  val avgParams = new Params
  val wordIndexer = new Indexer[String]
  val tagIndexer = new Indexer[String]
  val allowableTags = new HashMap[String,Array[Int]] // Word -> list of states/tags

  override def learnInit = {
    track("Indexing words and tags") {
      val map = new HashMap[String,HashSet[Int]]
      trainExamples.foreach { ex:Example =>
        foreach(ex.N, { i:Int =>
          wordIndexer.getIndex(ex.words(i))
          val a = tagIndexer.getIndex(ex.trueTags(i))
          if (opts.hmmUseTaggingDictionary)
            map.getOrElseUpdate(ex.words(i), new HashSet[Int]) += a
        })
      }
      logs("%s word types, %s tags", wordIndexer.size, tagIndexer.size)
      if (opts.hmmUseTaggingDictionary) {
        require (tagIndexer.size == K)
        map.keys.foreach { w =>
          allowableTags(w) = map(w).toList.sortWith(_<_).toArray
          //dbgs("%s: %s", w, allowableTags(w).map(tagIndexer.getObject(_)).mkString(" "))
        }
      }
    }
    super.learnInit
  }

  def initExample(ex:Example) = {
    opts.initType match {
      case Enums.InitType.wordtype =>
        // For each word type, choose a state to put it in
        // This isn't very good
        val word2state = new HashMap[String,Int]
        foreach(wordIndexer.size, { w:Int =>
          word2state(wordIndexer.getObject(w)) = opts.sampleRandom.nextInt(K)
        })
        foreach(ex.N, { i:Int =>
          ex.states(i) = word2state(ex.words(i))
        })
      case Enums.InitType.random =>
        foreach(ex.N, { i:Int =>
          ex.states(i) = {
            if (opts.hmmUseTaggingDictionary) {
              val l = allowableTags(ex.words(i))
              if (opts.sampleRandom.nextDouble < opts.initCutProb) l.head // All together (cut = 1: fewest patterns)
              else l(opts.sampleRandom.nextInt(l.size)) // Uniform (cut = 0: most patterns)
            }
            else {
              if (opts.sampleRandom.nextDouble < opts.initCutProb) 0 // All together (cut = 1: fewest patterns)
              else opts.sampleRandom.nextInt(K) // Uniform (cut = 0: most patterns)
            }
          }
        })
    }
  }

  ////////////////////////////////////////////////////////////

  class IterState extends super.IterState {
    def setState(ex:Example, i:Int, a:Int) : Boolean = {
      numSampledVars += 1
      ex.lastIterSampled(i) = iter
      if (ex.states(i) != a) {
        ex.states(i) = a
        numChangedVars += 1
        true
      }
      else
        false
    }

    override def processExampleToken(ex:Example) = {
      foreach(ex.N, { i:Int =>
        val prev_a = ex.getState(i-1)
        val next_a = ex.getState(i+1)
        val w = ex.words(i)
        def incrParams(dv:Double) = {
          val a = ex.states(i)
          if (i == 0) params.startCounts.incr(a, dv)
          else params.transitionCounts(prev_a).incr(a, dv)
          if (i < ex.N-1) params.transitionCounts(a).incr(next_a, dv)
          params.emissionCounts(a).incr(w, dv)
        }

        incrParams(-1)
        def score(a:Int) = {
          val prev_contrib = {
            if (i == 0) params.startAlpha/K + params.startCounts.getCount(a) 
            else params.transitionAlpha/K + params.transitionCounts(prev_a).getCount(a)
          }
          val next_contrib = {
            if (i == ex.N-1) 1.0
            else (params.transitionAlpha/K + params.transitionCounts(a).getCount(next_a)) /
                 (params.transitionAlpha + params.transitionCounts(a).totalCount)
          }
          val emit_contrib = {
            (params.emissionAlpha*emissionBaseProb(w) + params.emissionCounts(a).getCount(w)) /
            (params.emissionAlpha + params.emissionCounts(a).totalCount)
          }

          val weight = prev_contrib * next_contrib * emit_contrib
          if (temperature == 1) weight
          else math.pow(weight, 1.0/temperature)
        }

        val a = if (opts.hmmUseTaggingDictionary) {
          val l = allowableTags(w)
          if (opts.hardMoves) Utils.argmax(l.size, { q:Int => score(l(q)) })
          else l(Utils.normalizeAndSample(opts.sampleRandom, l.size, { q:Int => score(l(q)) }))
        }
        else {
          if (opts.hardMoves) Utils.argmax(K, score _)
          else Utils.normalizeAndSample(opts.sampleRandom, K, score _)
        }
        setState(ex, i, a)
        incrParams(+1)
      })
    }

    def existsAmbiguity(site:Site) = {
      !opts.hmmUseTaggingDictionary || allowableTags(site.ex.words(site.i)).size > 1
    }

    def newMoveState(site:Site) = {
      if (!existsAmbiguity(site)) None
      else {
        if (opts.hmmConsiderStatePairs) Some(new PairwiseMoveState(site))
        else Some(new AllMoveState(site))
      }
    }

    case class PairwiseMoveState(origSite:Site) extends super.MoveState {
      require (temperature == 1)
      require (opts.hmmConsiderStatePairs)

      val w = origSite.ex.words(origSite.i)
      val prev_a = origSite.ex.getState(origSite.i-1)
      val curr_a = origSite.ex.getState(origSite.i) // Consider other sites like this (configuration 0)
      val cand_a = { // Consider going to some other state (configuration 1)
        if (opts.hmmUseTaggingDictionary) {
          require (allowableTags(w).find(_ == curr_a) != None)
          val l = allowableTags(w).filter(_ != curr_a)
          l(opts.sampleRandom.nextInt(l.size))
        }
        else {
          val a = opts.sampleRandom.nextInt(HMMLearner.this.K-1)
          if (a >= curr_a) a + 1 else a
        }
      }
      require (curr_a != cand_a)
      val next_a = origSite.ex.getState(origSite.i+1)

      def c2a(c:Int) = if (c == 0) curr_a else cand_a

      val patterns = Pattern(prev_a, w, next_a) :: Nil

      def incrCounts(c:Int, d:Double) = {
        val a = c2a(c)
        if (prev_a != -1) params.transitionCounts(prev_a).incr(a, d) // From previous state
        else params.startCounts.incr(a, d)
        if (next_a != -1) params.transitionCounts(a).incr(next_a, d) // To next state
        params.emissionCounts(a).incr(w, d)
      }

      def getSite(o:PatternOccurrence) = {
        val ex = o.spot.ex
        val i = o.spot.i
        if (ex.states(i) == curr_a || ex.states(i) == cand_a) { // Must match the move we're trying to make
          o.spot.ex.spots(o.spot.i) match {
            case Some(spot) => if (spot.removed) None else Some(spot)
            case None => None
          }
        }
        else
          None
      }

      def numConfigurations = 2
      def getConfiguration(site:Site) = if (site.ex.getState(site.i) == curr_a) 0 else 1
      def setConfiguration(site:Site, c:Int) = setState(site.ex, site.i, c2a(c))

      def newScorer = new Scorer
      class ScoreInfo(a:Int) {
        val prevTransTop = {
          if (prev_a == -1)   params.startAlpha/K                      + params.startCounts.getCount(a)
          else                params.transitionAlpha/K                 + params.transitionCounts(prev_a).getCount(a)
        }
        val nextTransTop = {
          if (next_a == -1)   0.0 // Should not use this value in this case
          else                params.transitionAlpha/K                 + params.transitionCounts(a).getCount(next_a)
        }
        val nextTransBottom = params.transitionAlpha                   + params.transitionCounts(a).totalCount
        val emitTop         = params.emissionAlpha*emissionBaseProb(w) + params.emissionCounts(a).getCount(w)
        val emitBottom      = params.emissionAlpha                     + params.emissionCounts(a).totalCount
      }
      class Scorer extends AbstractScorer {
        val B = Array(new ScoreInfo(curr_a), new ScoreInfo(cand_a))
        val b_prevTransBottom = {
          if (prev_a == -1) params.startAlpha        + params.startCounts.totalCount
          else              params.transitionAlpha   + params.transitionCounts(prev_a).totalCount
        }

        def indepScore(c:Int) = throw fail("Not implemented")
        def localScore(k:Int, n:Int, m:Int) = throw fail("Not implemented")

        def computeScores(k:Int, n:Int) = {
          require (k == 2)
          // m is the number of sites assigned to cand_
          val weights = new BigDoubleArray(n+1)
          def mult(sign:Int, base:Double) = weights.multGenFact(sign, base, 1)
          def multRev(sign:Int, base:Double) = weights.multGenFactReverse(sign, base, 1)
          val c = k-1

          // Various cases based on prev_a and next_a
          if (prev_a == -1 && next_a == -1) { // -1 -> a -> -1
            mult(+1, B(c).prevTransTop) // Don't need bottom because it's constant
            multRev(+1, B(1-c).prevTransTop) // Don't need bottom because it's constant
          }
          else if (prev_a == -1) { // -1 -> a -> 1
            mult(+1, B(c).prevTransTop); mult(+1, B(c).nextTransTop); mult(-1, B(c).nextTransBottom)
            multRev(+1, B(1-c).prevTransTop); multRev(+1, B(1-c).nextTransTop); multRev(-1, B(1-c).nextTransBottom)
          }
          else if (next_a == -1) { // 0 -> a -> -1
            mult(+1, B(c).prevTransTop) // Don't need bottom because it's constant
            multRev(+1, B(1-c).prevTransTop) // Don't need bottom because it's constant
          }
          else if (prev_a == next_a) {
            if (cand_a == prev_a) { // 0 -> 0(cand) -> 0
              weights.multGenFact(+1, B(c).prevTransTop, 2); mult(-1, b_prevTransBottom+N) // Incremental bottom
            }
            else { // 0 -> a -> 0 (a != 0)
              mult(+1, B(c).prevTransTop); mult(+1, B(c).nextTransTop); mult(-1, B(c).nextTransBottom)
            }
            if (curr_a == prev_a) { // 0 -> 0(curr) -> 0
              weights.multGenFactReverse(+1, B(1-c).prevTransTop, 2); multRev(-1, b_prevTransBottom+N)
            }
            else { // 0 -> a -> 0 (a != 0)
              multRev(+1, B(1-c).prevTransTop); multRev(+1, B(1-c).nextTransTop); multRev(-1, B(1-c).nextTransBottom)
            }
          }
          else {
            if (cand_a == prev_a && curr_a == next_a) { // 0 -> 0(cand_a) -> 1 and 0 -> 1(curr_a) -> 1 (always N 0->1s)
              mult(+1, B(c).prevTransTop); mult(-1, b_prevTransBottom+N) // cand
              multRev(+1, B(1-c).nextTransTop); multRev(-1, B(1-c).nextTransBottom) // curr
            }
            else if (curr_a == prev_a && cand_a == next_a) { // 0 -> 0(curr_a) -> 1 and 0 -> 1(cand_a) -> 1
              mult(+1, B(c).nextTransTop); mult(-1, B(c).nextTransBottom) // cand
              multRev(+1, B(1-c).prevTransTop); multRev(-1, b_prevTransBottom+N) // curr
            }
            else { // 0 -> a -> 1 (a != 0, a != 1)
              mult(+1, B(c).prevTransTop); mult(+1, B(c).nextTransTop); mult(-1, B(c).nextTransBottom) // cand
              multRev(+1, B(1-c).prevTransTop); multRev(+1, B(1-c).nextTransTop); multRev(-1, B(1-c).nextTransBottom) // curr
            }
          }

          // Emissions
          mult(+1, B(1).emitTop); mult(-1, B(1).emitBottom) // cand
          multRev(+1, B(0).emitTop); multRev(-1, B(0).emitBottom) // curr

          mult(-1, 1); multRev(-1, 1)  // n choose k (ignore n!, which is a constant)

          val scores = weights.toLog
          if (opts.verbose >= 6) {
            dbgs("== %s -> %s|%s -> %s", prev_a, curr_a, cand_a, next_a)
            dbgs("A: %s", fmt1(scores))
          }
          scores
        }
      }
    }

    // Try to move all states at once (too hard)
    case class AllMoveState(origSite:Site) extends super.MoveState {
      require (temperature == 1)
      require (!opts.hmmConsiderStatePairs)
      require (!opts.hmmUseTaggingDictionary)

      val w = origSite.ex.words(origSite.i)
      val prev_a = origSite.ex.getState(origSite.i-1)
      val next_a = origSite.ex.getState(origSite.i+1)

      val patterns = Pattern(prev_a, w, next_a) :: Nil

      val c2a = map(K, { c:Int => c }) // Map from configuration c to state a
      def compute_a2c(a:Int) = c2a.indexOf(a)
      // Need to put next/prev states as the first two configurations so we can have a potential on their sum
      // FUTURE: further reordering to make prediction problem easier (sort by counts?)
      if (next_a != -1) Utils.swap(c2a, 1, compute_a2c(next_a))
      if (prev_a != -1) Utils.swap(c2a, 0, compute_a2c(prev_a))
      val next_c = compute_a2c(next_a)
      val prev_c = compute_a2c(prev_a)
      val a2c = map(K, { a:Int => compute_a2c(a) })

      require (prev_c == -1 || prev_c == 0)
      require (next_c == -1 || next_c == 0 || next_c == 1)
      require ((prev_c == next_c) == (prev_a == next_a))
      require ((prev_c == -1) == (prev_a == -1))
      require ((next_c == -1) == (next_a == -1))

      def incrCounts(c:Int, d:Double) = {
        val a = c2a(c)
        if (prev_a != -1) params.transitionCounts(prev_a).incr(a, d) // From previous state
        else params.startCounts.incr(a, d)
        if (next_a != -1) params.transitionCounts(a).incr(next_a, d) // To next state
        params.emissionCounts(a).incr(w, d)
      }

      def getSite(o:PatternOccurrence) = {
        o.spot.ex.spots(o.spot.i) match {
          case Some(spot) => if (spot.removed) None else Some(spot)
          case None => None
        }
      }

      def numConfigurations = HMMLearner.this.K
      def getConfiguration(site:Site) = a2c(site.ex.getState(site.i))
      def setConfiguration(site:Site, c:Int) = setState(site.ex, site.i, c2a(c))

      // This is not a very good proposal distribution
      def polyaProposeCounts(scorer:Scorer) : (Array[Int],Array[Int]=>Double) = {
        val scores = map(K, { c:Int => scorer.indepScore(c) })
        if (opts.verbose >= 1) dbgs("scores: %s", fmt1(scores))
        val alphas = Utils.expNormalizeOrFail_!(scores)
        foreach(K, { c:Int => alphas(c) = alphas(c)*opts.polyaConcentration + opts.polyaSmoothing/K })
        if (opts.verbose >= 1) dbgs("alphas: %s", fmt1(alphas))

        val proposalLogProb = { counts:Array[Int] =>
          // Polya distribution probability
          //Utils.sum(K, { c:Int => logGenFact(alphas(c), counts(c)) }) - logGenFact(Utils.sum(alphas), Utils.sum(counts))
          Utils.sum(K, { c:Int =>
            val a = logGenFact(alphas(c), counts(c))
            val b = logGenFact(Utils.sum(alphas), Utils.sum(counts))
            Utils.assertValid(a)
            Utils.assertValid(b)
            a-b
          })
        }

        val probs = fig.prob.Dirichlet.sample(opts.sampleRandom, alphas)
        val newCounts = fig.prob.SampleUtils.sampleMultinomial(opts.sampleRandom, N, probs)
        (newCounts, proposalLogProb)
      }

      def newScorer = new Scorer
      class ScoreInfo(a:Int) {
        val prevTransTop = {
          if (prev_a == -1)   params.startAlpha/K                      + params.startCounts.getCount(a)
          else                params.transitionAlpha/K                 + params.transitionCounts(prev_a).getCount(a)
        }
        /*params.write(10000, logs("%s", _))
        dbgs("%s %s %s", prev_a, a, prevTransTop)
        require (prevTransTop > 0)*/
        val nextTransTop = {
          if (next_a == -1)   0.0 // Should not use this value in this case
          else                params.transitionAlpha/K                 + params.transitionCounts(a).getCount(next_a)
        }
        val nextTransBottom = params.transitionAlpha                   + params.transitionCounts(a).totalCount
        val emitTop         = params.emissionAlpha*emissionBaseProb(w) + params.emissionCounts(a).getCount(w)
        val emitBottom      = params.emissionAlpha                     + params.emissionCounts(a).totalCount
      }
      class Scorer extends AbstractDynProgScorer {
        val B = map(K, { a:Int => new ScoreInfo(a) })
        val b_prevTransBottom = {
          if (prev_a == -1) params.startAlpha        + params.startCounts.totalCount
          else              params.transitionAlpha   + params.transitionCounts(prev_a).totalCount
        }

        //params.write(Integer.MAX_VALUE, logs("%s", _))

        override def localFeatures(k:Int, n:Int) = {
          val features = new ArrayBuffer[Double]
          features += k
          features += n
          features += b_prevTransBottom 
          B.foreach { b =>
            features += b.prevTransTop
            features += b.nextTransTop
            features += b.nextTransBottom
            features += b.emitTop
            features += b.emitBottom
          }
          features.toArray
        }

        def indepScore(c:Int) = {
          val a = c2a(c)

          // Various cases based on prev_a and next_a
          val transScore = {
            if (prev_a == -1 && next_a == -1) // -1 -> a -> -1
              logGenFact(B(a).prevTransTop, N) - logGenFact(b_prevTransBottom, N)
            else if (prev_a == -1) // -1 -> a -> 1
              (logGenFact(B(a).prevTransTop, N) - logGenFact(b_prevTransBottom, N)) + (logGenFact(B(a).nextTransTop, N) - logGenFact(B(a).nextTransBottom, N))
            else if (next_a == -1) // 0 -> a -> -1
              logGenFact(B(a).prevTransTop, N) - logGenFact(b_prevTransBottom, N)
            else if (prev_a == next_a) {
              if (a == prev_a) // 0 -> 0 -> 0
                logGenFact(B(a).prevTransTop, 2*N) - logGenFact(b_prevTransBottom, 2*N)
              else { // 0 -> a -> 0 (a != 0)
                (logGenFact(B(a).prevTransTop, N) - logGenFact(b_prevTransBottom, N)) + (logGenFact(B(a).nextTransTop, N) - logGenFact(B(a).nextTransBottom, N))
              }
            }
            else {
              if (a == prev_a) // 0 -> 0 -> 1
                logGenFact(B(a).prevTransTop, N) + logGenFact(B(a).nextTransTop, N) - logGenFact(b_prevTransBottom, 2*N)
              else if (a == next_a) // 0 -> 1 -> 1
                (logGenFact(B(a).prevTransTop, N) - logGenFact(b_prevTransBottom, N)) + (logGenFact(B(a).nextTransTop, N) - logGenFact(B(a).nextTransBottom, N))
              else // 0 -> a -> 1 (a != 0, a != 1)
                (logGenFact(B(a).prevTransTop, N) - logGenFact(b_prevTransBottom, N)) + (logGenFact(B(a).nextTransTop, N) - logGenFact(B(a).nextTransBottom, N))
            }
          }

          val emitScore = logGenFact(B(a).emitTop, N) - logGenFact(B(a).emitBottom, N)
          //dbgs("EMIT2: %s %s", transScore, emitScore)
          transScore + emitScore
        }

        override def computeInitScores = {
          val scores = localScores(1, N)
          /*val scores2 = map(N+1, { n:Int => localScore(1, n, n) })
          val scores3 = map(N+1, { n:Int => localScore(1, N, n) })
          if (Utils.distLinf(scores, scores2) > 1e-5) {
            dbgs("== %s -> %s -> %s", prev_a, c2a(0), next_a)
            dbgs("D: %s", fmt1(scores))
            dbgs("E: %s", fmt1(scores2))
            dbgs("F: %s", fmt1(scores3))
          }*/
          scores
        }

        override def localScores(k:Int, n:Int) = {
          val a = c2a(k-1)
          val weights = new BigDoubleArray(n+1)
          def mult(sign:Int, base:Double) = weights.multGenFact(sign, base, 1)

          // Various cases based on prev_a and next_a
          if (prev_a == -1 && next_a == -1) // -1 -> a -> -1
            mult(+1, B(a).prevTransTop) // Don't need bottom because it's constant
          else if (prev_a == -1) { // -1 -> a -> 1
            mult(+1, B(a).prevTransTop); mult(+1, B(a).nextTransTop); mult(-1, B(a).nextTransBottom)
          }
          else if (next_a == -1) // 0 -> a -> -1
            mult(+1, B(a).prevTransTop) // Don't need bottom because it's constant
          else if (prev_a == next_a) {
            if (a == prev_a) { // 0 -> 0 -> 0
              weights.multGenFact(+1, B(a).prevTransTop, 2); weights.multExp(-logGenFact(b_prevTransBottom, N)); mult(-1, b_prevTransBottom+N)
            }
            else { // 0 -> a -> 0 (a != 0)
              mult(+1, B(a).prevTransTop); mult(+1, B(a).nextTransTop); mult(-1, B(a).nextTransBottom)
            }
          }
          else {
            if (a == prev_a) { // 0 -> 0 -> 1
              mult(+1, B(a).prevTransTop); weights.multExp(-logGenFact(b_prevTransBottom, N)); mult(-1, b_prevTransBottom+N) // Transition 0 -> 1 covered by next case
            }
            else if (a == next_a) { // 0 -> 1 -> 1
              weights.multExp(logGenFact(B(a).prevTransTop, n)); mult(+1, B(a).nextTransTop); mult(-1, B(a).nextTransBottom) // Note n not m (counts 0 -> 1 above too)
            }
            else { // 0 -> a -> 1 (a != 0, a != 1)
              mult(+1, B(a).prevTransTop); mult(+1, B(a).nextTransTop); mult(-1, B(a).nextTransBottom)
            }
          }

          mult(+1, B(a).emitTop)
          mult(-1, B(a).emitBottom)

          val scores = weights.toLog
          /*val scores2 = map(n+1, { m:Int => localScore2(k, n, m) })
          if (Utils.distLinf(scores, scores2) > 1e-5) {
            dbgs("== %s -> %s -> %s", prev_a, a, next_a)
            dbgs("A: %s", fmt1(scores))
            dbgs("B: %s", fmt1(scores2))
          }*/
          scores
        }

        def localScore(k:Int, n:Int, m:Int) = {
          //val diff = localScore1(k, n, m) - localScore2(k, n, m)
          //require (diff.abs < 1e-30, k + " " + n + " " + m + " : " + diff)
          localScore2(k, n, m)
        }

        def localScore2(k:Int, n:Int, m:Int) = {
          val a = c2a(k-1)

          // Various cases based on prev_a and next_a
          val transScore = {
            if (prev_a == -1 && next_a == -1) // -1 -> a -> -1
              logGenFact(B(a).prevTransTop, m) // Don't need bottom because it's constant
            else if (prev_a == -1) // -1 -> a -> 1
              logGenFact(B(a).prevTransTop, m) + (logGenFact(B(a).nextTransTop, m) - logGenFact(B(a).nextTransBottom, m))
            else if (next_a == -1) // 0 -> a -> -1
              logGenFact(B(a).prevTransTop, m) // Don't need bottom because it's constant
            else if (prev_a == next_a) {
              if (a == prev_a) // 0 -> 0 -> 0
                logGenFact(B(a).prevTransTop, 2*m) - logGenFact(b_prevTransBottom, N+m)
              else { // 0 -> a -> 0 (a != 0)
                logGenFact(B(a).prevTransTop, m) + (logGenFact(B(a).nextTransTop, m) - logGenFact(B(a).nextTransBottom, m))
              }
            }
            else {
              if (a == prev_a) // 0 -> 0 -> 1
                logGenFact(B(a).prevTransTop, m) - logGenFact(b_prevTransBottom, N+m) // Transition 0 -> 1 covered by next case
              else if (a == next_a) // 0 -> 1 -> 1
                logGenFact(B(a).prevTransTop, n) + (logGenFact(B(a).nextTransTop, m) - logGenFact(B(a).nextTransBottom, m)) // Note n not m (counts 0 -> 1 above too)
              else // 0 -> a -> 1 (a != 0, a != 1)
                logGenFact(B(a).prevTransTop, m) + (logGenFact(B(a).nextTransTop, m) - logGenFact(B(a).nextTransBottom, m))
            }
          }

          val emitScore = logGenFact(B(a).emitTop, m) - logGenFact(B(a).emitBottom, m)
          //dbgs("EMIT2: %s %s", transScore, emitScore)
          transScore + emitScore
        }

        // Don't use because too slow
        /*def startTop(a:Int, m:Int)        = logGenFact(params.startAlpha/K                      + params.startCounts.getCount(a),         m)
        def transTop(a:Int, b:Int, m:Int) = logGenFact(params.transitionAlpha/K                 + params.transitionCounts(a).getCount(b), m)
        def transBottom(a:Int, m:Int)     = logGenFact(params.transitionAlpha                   + params.transitionCounts(a).totalCount,  m)
        def emitTop(a:Int, m:Int)         = logGenFact(params.emissionAlpha*emissionBaseProb(w) + params.emissionCounts(a).getCount(w),   m)
        def emitBottom(a:Int, m:Int)      = logGenFact(params.emissionAlpha                     + params.emissionCounts(a).totalCount,    m)
        def localScore1(k:Int, n:Int, m:Int) = {
          val a = c2a(k-1)

          // Various cases based on prev_a and next_a
          val transScore = {
            if (prev_a == -1 && next_a == -1) // -1 -> a -> -1
              startTop(a, m) // Don't need bottom because it's constant
            else if (prev_a == -1) // -1 -> a -> 1
              startTop(a, m) + (transTop(a, next_a, m) - transBottom(a, m))
            else if (next_a == -1) // 0 -> a -> -1
              transTop(prev_a, a, m) // Don't need bottom because it's constant
            else if (prev_a == next_a) {
              if (a == prev_a) // 0 -> 0 -> 0
                transTop(prev_a, a, 2*m) - transBottom(prev_a, N+m)
              else { // 0 -> a -> 0 (a != 0)
                //dbgs("HERE1: %s %s %s", transTop(prev_a, a, m), transTop(a, next_a, m), transBottom(a, m))
                transTop(prev_a, a, m) + (transTop(a, next_a, m) - transBottom(a, m))
              }
            }
            else {
              if (a == prev_a) // 0 -> 0 -> 1
                transTop(prev_a, a, m) - transBottom(prev_a, N+m) // Transition 0 -> 1 covered by next case
              else if (a == next_a) // 0 -> 1 -> 1
                transTop(prev_a, a, n) + (transTop(a, next_a, m) - transBottom(a, m)) // Note n not m (counts 0 -> 1 above too)
              else // 0 -> a -> 1 (a != 0, a != 1)
                transTop(prev_a, a, m) + (transTop(a, next_a, m) - transBottom(a, m))
            }
          }

          val emitScore = emitTop(a, m) - emitBottom(a, m)
          //dbgs("EMIT1: %s %s", transScore, emitScore)
          transScore + emitScore
        }*/
      }
    }
  }
  def newIterState = new IterState

  def stateLogProb : Double = {
    var logProb = 0.0
    logProb += margLogProb(params.startCounts, params.startAlpha, transitionBaseLogProb _)
    foreach(K, { a:Int =>
      logProb += margLogProb(params.transitionCounts(a), params.transitionAlpha, transitionBaseLogProb _)
      logProb += margLogProb(params.emissionCounts(a), params.emissionAlpha, emissionBaseLogProb _)
    })
    logProb
  }
  override def stateAccuracy = {
    if (opts.hmmUseTaggingDictionary) stateIdentityMappedAccuracy
    else stateGreedyMappedAccuracy
  }

  def stateIdentityMappedAccuracy = {
    // Compute a greedy mapping 
    val counts = map(K, { a:Int => new CounterMap[String] })

    trainExamples.foreach { ex =>
      foreach(ex.N, { i:Int =>
        counts(ex.states(i)).incr(ex.trueTags(i), 1)
      })
    }

    if (opts.verbose >= 0) {
      // Print confusion matrix: predicted states are rows, true tags are columns
      val tags = tagIndexer.getObjects.toArray[String](new Array[String](0))
      val table = {
        Array(Array("") ++ tags) ++
        map(K, { a:Int =>
          Array(tags(a)) ++ tags.map { tag:String => counts(a).getCount(tag).toString }
        })
      }
      Utils.formatTable(table, { (r:Int,c:Int) => if (c == 0) -1 else if (r == 0) 0 else 1 }).foreach(logs("%s", _))
    }

    // Greedily map each predicted state k to the best true tag t
    Utils.sum(K, { a:Int => counts(a).getCount(tagIndexer.getObject(a)) }) /
    Utils.sum(K, { a:Int => counts(a).totalCount })
  }

  def stateGreedyMappedAccuracy = {
    // Compute a greedy mapping 
    val counts = map(K, { a:Int => new CounterMap[String] })
    val tagSet = new HashSet[String]

    trainExamples.foreach { ex =>
      foreach(ex.N, { i:Int =>
        counts(ex.states(i)).incr(ex.trueTags(i), 1)
        tagSet += ex.trueTags(i)
      })
    }

    if (opts.verbose >= 0) {
      // Print confusion matrix: predicted states are rows, true tags are columns
      val tags = tagSet.toList.sortWith(_ < _)
      val table = {
        Array(Array("") ++ tags) ++
        map(K, { a:Int =>
          Array(a+"("+counts(a).argmaxAndMax._1.getOrElse("-")+")") ++ tags.map { tag:String => counts(a).getCount(tag).toString }
        })
      }
      Utils.formatTable(table, { (r:Int,c:Int) => if (c == 0) -1 else if (r == 0) 0 else 1 }).foreach(logs("%s", _))
    }

    // Greedily map each predicted state k to the best true tag t
    Utils.sum(K, { a:Int => counts(a).argmaxAndMax._2 }) /
    Utils.sum(K, { a:Int => counts(a).totalCount })
  }

  // Uniform over words and transitions
  def emissionBaseProb(w:String) = 1.0/wordIndexer.size
  def emissionBaseLogProb(w:String) = math.log(emissionBaseProb(w))
  def transitionBaseProb(a:Int) = 1.0/K
  def transitionBaseLogProb(a:Int) = math.log(transitionBaseProb(a))

  override def repCheck = track("repCheck") {
    super.repCheck
    def change(dv:Double) = trainExamples.foreach(_.incrParams(dv))
    change(-1)
    require (params.startCounts.numNonZero == 0)
    foreach(K, { a:Int =>
      require (params.transitionCounts(a).numNonZero == 0)
      require (params.emissionCounts(a).numNonZero == 0)
    })
    change(+1)
  }

  override def stateStr(ex:Example) = ex.states.mkString(" ")
  def getInferenceStats(ex:Example) = ex.counts.flatMap{x => x.map(_.toDouble / ex.totalCount)}
}
