package typesampling

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import java.io.File

class MainOptions {
  import tea.OptionTypes._

  @Option var domainType = Enums.DomainType.java
  @Option var fileExt : String = null

  // Input
  @Option var trainBasePaths = Array[String]()
  @Option var testBasePaths = Array[String]()
  @Option var trainMaxPaths = Integer.MAX_VALUE
  @Option var trainMaxNodes = Integer.MAX_VALUE
  @Option var testMaxPaths = Integer.MAX_VALUE
  @Option var testMaxNodes = Integer.MAX_VALUE

  // Generation
  @Option var genRandom = new Random(1)
  @Option var generate = false
}

object Main extends Runnable {
  import tea.OptionTypes._
  @OptionSet(name="main") val mainOptions = new MainOptions
  @OptionSet(name="learn") val learnOptions = new LearnOptions
  @OptionSet(name="toy") val toyOptions = new ToyOptions
  @OptionSet(name="wsj") val wsjOptions = new WSJOptions
  @OptionSet(name="words") val wordsOptions = new WordsOptions

  import mainOptions._

  def getPaths(basePaths:Seq[String], ext:String, maxPaths:Int) = {
    val paths = new ArrayBuffer[String]
    def recursiveAdd(path:File) : Unit = { 
      if (paths.size < maxPaths) {
        if (path.isDirectory)
          path.listFiles.foreach(recursiveAdd(_))
        else {
          if (path.toString.endsWith("."+ext)) paths += path.toString
        }
      }
    }
    basePaths.foreach { sourcePath:String => recursiveAdd(new File(sourcePath)) }
    paths.toArray
  }

  def getDomain = {
    domainType match {
      case Enums.DomainType.toy => new ToyDomain(toyOptions)
      case Enums.DomainType.wsj => new WSJDomain(wsjOptions)
      case Enums.DomainType.words => new WordsDomain(wordsOptions)
    }
  }
  def getLearner = {
    learnOptions.learnerType match {
      case Enums.LearnerType.ptsg => new ptsg.PTSGLearner(learnOptions)
      case Enums.LearnerType.flatseg => new flatseg.FlatSegLearner(learnOptions)
      case Enums.LearnerType.flatseg2 => new flatseg2.FlatSegLearner2(learnOptions)
      case Enums.LearnerType.hmm => new hmm.HMMLearner(learnOptions)
    }
  }

  def readNodes(paths:Array[String], maxNodes:Int) = {
    val domain = getDomain
    val nodes = new ArrayBuffer[Node]
    foreach(paths.size, { i:Int =>
      val path = paths(i)
      if (nodes.size < maxNodes) {
        track("Path %s/%s: %s", i, paths.size, path) {
          nodes ++= domain.readNodes(path, maxNodes-nodes.size)
        }
      }
    })
    nodes.toArray
  }

  def run = {
    val domain = getDomain

    // Read input
    val fileExt = if (!Utils.empty_?(mainOptions.fileExt)) mainOptions.fileExt else domain.fileExt
    val trainPaths = getPaths(trainBasePaths, fileExt, trainMaxPaths)
    val testPaths = getPaths(testBasePaths, fileExt, testMaxPaths)

    val (trainNodes, testNodes) = {
      track("Reading %s train paths, %s test paths", trainPaths.size, testPaths.size) {
        (readNodes(trainPaths, trainMaxNodes), readNodes(testPaths, testMaxNodes))
      }
    }

    // Learn a model
    val learner = getLearner
    track("Learn") {
      learner.setTrainNodes(trainNodes)
      learner.setTestNodes(testNodes)
      learner.learn
    }

    // Evaluate it on new data
    /*Utils.track_printAll("Evaluate") {
      learner.evaluate("train", trainNodes)
      learner.evaluate("test", testNodes)
    }*/

    // Generate from the model
    if (generate) track("Generate") {
      val node = learner.generate(genRandom)
      Utils.writeLines(Execution.getFile("generated.node"), { puts:(String=>Any) => puts(node.toString) })
      Utils.writeLines(Execution.getFile("generated.out"), { puts:(String=>Any) => domain.write(node, puts) })
    }
  }

  def main(args:Array[String]) = {
    Execution.run(args, "Main", this)
  }
}
