package typesampling

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid}
import tea.Utils.{track,begin_track,end_track,logs,dbgs,error}

import java.util.{Set,ArrayList,Random}
import java.io.File

import edu.berkeley.nlp.ling._
import edu.berkeley.nlp.parser._

/*
Tree representations:
 (A) WSJ trees (with traces)
 (B) WSJ transformed trees (traces removed)
 (C) annotated WSJ trees (binarized, parent annotated, flattened unary chains)

Read in gold trees as (A).
Evaluate on (B).
Train on (C).
Predictions from decoder are in (C).
*/

class WSJOptions {
  import tea.OptionTypes._
  @Option var binarize = false
  @Option var binarizeOrder = 0
  @Option var parentAnnotate = false
  @Option var flattenUnaryChains = true
  @Option var annotateUnaryChains = false
}

class WSJDomain(options:WSJOptions) extends Domain {
  def fileExt = "mrg"

  type Tree = edu.berkeley.nlp.ling.Tree[String]

  def write(node:Node, puts:String=>Any) = node.write(puts)

  def readNodes(path:String, maxNodes:Int) = {
    val nodes = new ArrayBuffer[Node]
    WSJDomain.foreachTree(path, maxNodes, true, { _tree:Tree =>
      var tree = _tree
      if (tree.getLabel == "ROOT" && tree.getChildren.size == 1)
        tree = tree.getChildren.get(0)
      //logs("%s", TreeUtils.render(tree))
      if (options.parentAnnotate) tree = TreeUtils.parentAnnotate(tree)
      //logs("%s", TreeUtils.render(tree))
      if (options.binarize) tree = TreeUtils.headBinarize(tree, options.binarizeOrder)
      //logs("%s", TreeUtils.render(tree))
      if (options.flattenUnaryChains) tree = TreeUtils.flattenUnaryChains(tree, options.annotateUnaryChains)
      //logs("%s", TreeUtils.render(tree))
      nodes += toNode(tree)
    })
    nodes
  }

  def toNode(tree:Tree) : Node = {
    val context = PrimitiveContext(tree.getLabel)
    val children = tree.getChildren
    if (tree.isPreTerminal) {
      require (children.size == 1)
      PrimitiveNode(context, children.get(0).getLabel)
    }
    else {
      require (children.size == 2)
      CompositeNode(context, toNode(children.get(0)), toNode(children.get(1)))
    }
  }
}

object WSJDomain {
  type Tree = edu.berkeley.nlp.ling.Tree[String]

  def foreachTree(path:String, maxTrees:Int, transform:Boolean, f:(Tree => Any)) = {
    val treeTransformer = new Trees.StandardTreeNormalizer
    val in = fig.basic.IOUtils.openIn(path)
    val treeIterator = new Trees.PennTreeReader(in)
    var n = 0
    while(n < maxTrees && treeIterator.hasNext) {
      var tree = treeIterator.next
      //logs("BEFORE TRANSFORM: %s", TreeUtils.render(tree))
      if (transform) tree = treeTransformer.transformTree(tree)
      //logs("AFTER TRANSFORM: %s", TreeUtils.render(tree))
      f(tree)
      n += 1
    }
    in.close
  }

  def readTrees(path:String, transform:Boolean) = {
    val trees = new ArrayBuffer[Tree]
    foreachTree(path, Integer.MAX_VALUE, transform, { tree:Tree =>
      trees += tree
    })
    trees.toArray
  }
}

////////////////////////////////////////////////////////////

class ParsingEvaluator {
  val eval = new EnglishPennTreebankParseEvaluator.LabeledConstituentEval[String](
      java.util.Collections.singleton("ROOT"),
      new java.util.HashSet[String](java.util.Arrays.asList(Array("''", "``", ".", ":", ",") : _*)))
  var numSentences = 0

  def add(trueTree:Tree[String], _predTree:Tree[String], pwOption:Option[java.io.PrintWriter]) = {
    numSentences += 1
    val predTree = TreeUtils.polishTree(_predTree)
    // predTree could be (ROOT), in which case it's unparsable.
    // If claimed to have parsed something and it doesn't match the sentence, then raise a red flag.
    //println(predTree.getChildren.size)
    //println(predTree.getYield.size)
    if (predTree.getYield.size != 1 && trueTree.getYield.size != predTree.getYield.size) {
      throw new RuntimeException("Mismatch in number of leaves (these might be parse trees for different sentences):\n"+
        TreeUtils.render(trueTree) + "\n" + TreeUtils.render(predTree))
    }
    pwOption match { 
      case Some(pw) =>
        pw.println("=== Sentence "+numSentences+" ===")
        pw.println(TreeUtils.render(trueTree))
        pw.println(TreeUtils.render(predTree))
        eval.evaluate(predTree, trueTree, pw)
      case None =>
        eval.evaluate(predTree, trueTree)
    }
  }

  def display(pw:java.io.PrintWriter) = eval.display(true, pw)
  def dumpStats(puts:String=>Any) = {
    puts("numSentences\t"+numSentences)
    puts("precision\t"+eval.getPrecision)
    puts("recall\t"+eval.getRecall)
    puts("f1\t"+eval.getF1)
    puts("exactMatch\t"+eval.getExactMatch)
  }
}

object ParsingEvaluatorMain {
  def main(args:Array[String]) = {
    if (args.length < 2) {
      println("Usage: <in: true trees file> <in: pred trees file> [out: eval file] [out: detailed eval file]")
      sys.exit(1)
    }
    fig.basic.LogInfo.msPerLine = 0

    val eval = new ParsingEvaluator
    val trueTrees = WSJDomain.readTrees(args(0), true)
    val predTrees = WSJDomain.readTrees(args(1), false)
    val evalPath = if (args.length > 2) args(2) else "/dev/stdout"
    val detailedOutOption = if (args.length > 3) Some(fig.basic.IOUtils.openOutHard(args(3))) else None

    if (trueTrees.size != predTrees.size) {
      logs("WARNING: "+trueTrees.size+" true trees, "+predTrees.size+" pred trees")
    }
    foreach(trueTrees.size min predTrees.size, { i:Int =>
      eval.add(trueTrees(i), predTrees(i), detailedOutOption)
    })

    detailedOutOption match {
      case Some(pw) =>
        eval.display(pw)
        pw.close
      case None =>
    }

    Utils.writeLines(evalPath, eval.dumpStats _)
    eval.dumpStats(println)
  }
}

////////////////////////////////////////////////////////////

object AbstractRareWords {
  def main(args:Array[String]) = {
    if (args.length < 3) {
      println("Usage: <in: lexicon file> <in: sentences> <out: new sentences>")
      sys.exit(1)
    }
    val Array(lexiconPath, inPath, outPath) = args

    val lexicon = new Lexicon
    track("Reading lexicon from "+lexiconPath) {
      Utils.foreachLine(lexiconPath, { line:String =>
        val Array(word, count) = line.split("\\s+") // <word> <count>
        lexicon.add(word, count.toDouble)
        true
      })
    }

    track("Converting "+inPath+" to "+outPath) {
      Utils.writeLines(outPath, { puts:(String=>Any) =>
        Utils.foreachLine(inPath, { line:String =>
          val words = line.split("\\s+")
          val newWords = (0 to words.size-1).map { i:Int => lexicon.abstractIfRare(words(i), i) }
          puts(newWords.mkString(" "))
          true
        })
      })
    }
  }
}

////////////////////////////////////////////////////////////

class Lexicon {
  val rareCutoff = 1
  private val counts = new CounterMap[String]

  // At training time
  def add(word:String) : Unit = add(word, 1)
  def add(word:String, count:Double) = counts.incr(word, count)

  def abstractIfRare(word:String, loc:Int) = {
    if (counts.getCount(word) > rareCutoff) word
    else RareWordSignature.getSignature(word, loc)
  }

  def buildLexicon(trainNodes:Array[Node]) = {
    track("Building lexicon") {
      trainNodes.foreach { rootNode:Node =>
        rootNode.foreachNode { node:Node =>
          node match {
            case node:PrimitiveNode => add(node.value.asInstanceOf[String])
            case _ =>
          }
        }
      }
      logs("%s word tokens, %s word types", counts.totalCount, counts.numNonZero)
    }
  }

  def abstractRareWords(node:Node) : Node = {
    var i = 0
    def convert(node:Node) : Node = {
      node match {
        case node:PrimitiveNode =>
          val newNode = new PrimitiveNode(node.context, abstractIfRare(node.value.asInstanceOf[String], i))
          i += 1
          newNode
        case node:CompositeNode => new CompositeNode(node.context, convert(node.left), convert(node.right))
      }
    }
    convert(node)
  }

  def write(path:String) = {
    Utils.writeLines(path, counts.write(Integer.MAX_VALUE, false, "", _))
  }
}
